#!/usr/bin/env python
"""HiJitReg registers color CCDs to corresponding red CCDs by using the
ISIS tool hijitreg to perform a deconvolution of jittered image data.

This program corrects for spacecraft jitter and prepares
images for coregistration. Using the ISIS program hijitreg it creates
a reference grid of control points (extension .control.pvl) from
the RED product, and then attempts to locate those points within
the BG and IR products.

The configuration file for HiJitReg (HiJitReg.conf) describes the control
point grid density (currently it's a 4-column, 200-line grid) and
correlation tolerance. It also specifies the sizes of the search
window and pattern window. The pattern window "walks" through the
search window in order to locate the local maximum. The calculated
translation is then recorded in the control net file for slithering.

Some adaptability is built-in to HiJitReg. It will add columns if
a channel is missing, and will triple the grid point density if
less than 25% of the points register on the first pass. It will
also increase the size of the search window if more than two points
have the edge of their pattern box close to or beyond the edge of
the search box ("edgy" points).

It also uses a smoothing algorithm to ignore points that are
out-of-bounds or are poorly registered. A JitterPlot is created
showing the results.

HiJigReg also works on one or both of the "color" sets:

- RED4 - BG12 - IR10
- RED5 - BG13 - IR11

HiJitReg must be run after HiColorInit, and it creates a .hislither.pvl
file which is then submitted to the HiSlither_Pipeline.

Data Flow
---------
Input Products:

- RED4 and 5 ``balance.cub`` files which are the result of HiccdStitch.
- BG and IR ``precolor.cub`` files which are the result of HiColorInit.

Output Products:

- creates regdef.pvl and flat.tab and control.pvl files for each
    BG and IR cube provided.

"""

# Copyright 2004-2020, Arizona Board of Regents on behalf of the Lunar and
# Planetary Laboratory at the University of Arizona.
#   - Orignal Perl program.
#
# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#   - Elements of this Python program are are based on the original Perl
#     but the logic here is rewritten from scratch to emulate functionality.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This program is based on HiColor version 5.4.2 (2020/02/14),
# This program is based on these Perl programs:
# - HiJitReg.pm ($Revision: 1.58 $ $Date: 2020/04/28 16:56:16 $)
# - JitStats ($Revision: 1.11 $ $Date: 2020/04/28 16:56:16 $)
# - JitStats.pm ($Revision: 1.17 $ $Date: 2020/02/14 22:46:49 $)
# by Guy McArthur as an employee of the University of Arizona.

import argparse
import collections
import csv
import itertools
import logging
import math
import os
import pkg_resources
import re
import statistics
from datetime import datetime
from pathlib import Path

import pvl

import kalasiris as isis
import hiproc.util as util
import hiproc.HiColorInit as hicolor

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        parents=[util.parent_parser()],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "-c",
        "--conf",
        required=False,
        type=argparse.FileType('r'),
        default=pkg_resources.resource_stream(
            __name__,
            'data/HiJitReg.conf'
        ),
        help="Path to the HiJitReg config file.  Defaults to "
             "HiJitReg.conf distributed with the library."
    )
    parser.add_argument(
        "cubes",
        metavar="balance.cub and balance.precolor.cub files",
        nargs="+",
        help="Either one or both sets of RED .balance.cub  and IR/BG "
             ".balance.precolor.cub files. However, that's tedious to type,"
             "so you could just type in *.balance*cub here, and the "
             "program will sort out what it needs."
    )
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    with util.main_exceptions(args.verbose):
        successful_ccds = HiJitReg(
            args.cubes,
            pvl.load(args.conf),
            keep=args.keep
        )

    print("Successful CCDs are:")
    for c in successful_ccds:
        print("\t{}".format(str(c)))
    return


class JitterCube(hicolor.HiColorCube, collections.abc.MutableMapping):
    """A class for collecting and analyzing jitter statistics."""

    def __init__(
        self,
        arg,
        config=pvl.load(
            pkg_resources.resource_stream(
                __name__,
                'data/HiJitReg.conf'
            ),
        ),
        matchccd=None,
    ):
        if isinstance(arg, hicolor.HiColorCube):
            super().__init__(arg.path)
        else:
            super().__init__(arg)

        # self = copy.deepcopy(hi_color_cube)
        self.dictionary = dict()
        # HiColorCube isn't a dictionary, but we'll give access to its
        #   members as if it were:
        self.dictionary["bin"] = self.bin
        self.dictionary["tdi"] = self.tdi
        self.dictionary["lines"] = self.lines
        self.dictionary["samps"] = self.samps

        self.dictionary["CanSlither"] = False

        self.IgnoredPoints = set()

        # Just assuming that all of these will be in the self.dictionary
        # self.RegisterCount = None
        # self.AvgSampleOffset = None
        # self.AvgLineOffset = None
        # self.STDSampleOffset = None
        # self.STDLineOffset = None
        # self.SuspectCount = None
        # self.MatchedCount = None
        # self.RegisterCount = None
        # self.SearchSamples = None
        # self.SearchLines = None
        # self.EdgyCount = None
        # self.MatchedLineCount = None
        # self.Tolerance = None
        # self.Columns = None
        # self.Rows = None
        # self.canSlither = None
        # self.PatternSamples = None
        # self.PatternLines = None
        # self.SearchSamples = None
        # self.SearchLines = None

        if isinstance(config, (Path, str)):
            self.conf = pvl.load(str(config))
        elif isinstance(config, collections.abc.Mapping):
            self.conf = config
        else:
            raise TypeError(
                f"The value for *config* was neither an os.PathLike nor a "
                f"Mapping object, it was a {type(config)}"
            )

        if matchccd is None:
            self.matchccd = hicolor.CCD_Corresponence[self.get_ccd()]
        else:
            self.matchccd = matchccd

        sm = self.conf["Smoothing"]
        self.dictionary["ExcludeLimit"] = sm["Exclude_Limit"]
        self.dictionary["BadnessLimit"] = sm["Badness_Limit"]
        self.dictionary["BoxcarLength"] = sm["Boxcar_Length"]

        self.cnet_path = self.get_cnet_path(self)
        self.regdef_path = self.get_regdef_path(self)
        self.flattab_path = self.get_flattab_path(self)

    def __getitem__(self, key):
        return self.dictionary[key]

    def __setitem__(self, key, value):
        self.dictionary[key] = value

    def __delitem__(self, key):
        del self.dictionary[key]

    def __iter__(self):
        return iter(self.dictionary)

    def __len__(self):
        return len(self.dictionary)

    @staticmethod
    def get_pair_name(cube, matchccd=None):
        if matchccd is None:
            if hasattr(cube, "matchccd"):
                matchccd = cube.matchccd
            else:
                matchccd = hicolor.CCD_Corresponence[cube.get_ccd()]
        pair_name = "{}_{}-{}".format(
            str(cube.get_obsid()), matchccd, cube.get_ccd()
        )
        return pair_name

    @staticmethod
    def _get_path(cube, suffix):
        pair = JitterCube.get_pair_name(cube)
        return cube.path.parent / (pair + suffix)

    @staticmethod
    def get_cnet_path(cube):
        return JitterCube._get_path(cube, ".control.pvl")

    @staticmethod
    def get_regdef_path(cube):
        return JitterCube._get_path(cube, ".regdef.pvl")

    @staticmethod
    def get_flattab_path(cube):
        return JitterCube._get_path(cube, ".flat.tab")

    def reset(self):
        self.IgnoredPoints.clear()
        self.parseRegDefs(self.regdef_path)
        self.parseFlatTab(self.flattab_path)
        self.parseCNetPVL(self.cnet_path)

    def parseRegDefs(self, path=None):
        """Parse the register definition file to obtain the search and pattern
        sizes."""
        if path is None:
            path = self.regdef_path

        p = pvl.load(str(path))
        self["PatternSamples"] = p["AutoRegistration"]["PatternChip"][
            "Samples"
        ]
        self["PatternLines"] = p["AutoRegistration"]["PatternChip"]["Lines"]
        self["SearchSamples"] = p["AutoRegistration"]["SearchChip"]["Samples"]
        self["SearchLines"] = p["AutoRegistration"]["SearchChip"]["Lines"]
        return

    def parseFlatTab(self, path=None):
        """Parses the flat file to obtain jitter registration result
        statistics."""
        if path is None:
            path = self.flattab_path

        with open(path, "r") as f:
            flat = f.read()

            match = re.search(r"#\s+Line Spacing:\s+(\S+)", flat)
            self["LineSpacing"] = float(match.group(1))

            match = re.search(r"#\s+Columns, Rows:\s+(\d+)\s+(\d+)", flat)
            self["Columns"] = int(match.group(1))
            self["Rows"] = int(match.group(2))

            match = re.search(r"#\s+Corr. Tolerance:\s+(\S+)", flat)
            self["Tolerance"] = float(match.group(1))

            match = re.search(r"#\s+Total Registers:\s+(\d+) of (\S+)", flat)
            self["MatchedCount"] = int(match.group(1))
            self["RegisterCount"] = int(match.group(2))

            match = re.search(r"#\s+Number Suspect:\s+(\S+)", flat)
            self["SuspectCount"] = int(match.group(1))

            match = re.search(
                r"#\s+Average Sample Offset:\s+(\S+)\s+StdDev:\s+(\S+)", flat
            )
            self["AvgSampleOffset"] = float(match.group(1))
            self["STDSampleOffset"] = float(match.group(2))

            match = re.search(
                r"#\s+Average Line Offset:\s+(\S+)\s+StdDev:\s+(\S+)", flat
            )
            self["AvgLineOffset"] = float(match.group(1))
            self["STDLineOffset"] = float(match.group(2))

            dialect = csv.Dialect
            dialect.delimiter = " "
            dialect.skipinitialspace = True
            dialect.quoting = csv.QUOTE_NONE
            dialect.lineterminator = "\n"

            reader = csv.DictReader(
                itertools.filterfalse(
                    lambda x: x.startswith("#") or x.isspace() or len(x) == 0,
                    flat.splitlines(),
                ),
                dialect=dialect,
            )

            if "EdgyCount" not in self:
                self["EdgyCount"] = 0

            lineCount = 0
            for row in reader:
                # how many pixels in x is the edge of the pattern box from
                # the reg point
                deltaSamp = (
                    abs(float(row["RegSamp"]) - int(row["MatchSamp"]))
                    + self["PatternSamples"] / 2
                )
                # how many pixels in y is the edge of the pattern box from
                # the reg point
                deltaLine = (
                    abs(float(row["RegLine"]) - int(row["MatchLine"]))
                    + self["PatternLines"] / 2
                )

                # if the edge of the pattern box is more than two pixels away
                # from the search box, increment the count of marginal
                # control points
                if (deltaSamp > (self["SearchSamples"] / 2 - 2)) or (
                    deltaLine > (self["SearchLines"] / 2 - 2)
                ):
                    self["EdgyCount"] += 1

                    logger.info(
                        "Marginal register {} lines, {} samples to "
                        "edge.".format(
                            deltaLine - self["SearchLines"] / 2,
                            deltaSamp - self["SearchSamples"] / 2,
                        )
                    )
                lineCount += 1
            self["MatchedLineCount"] = lineCount
        return

    def parseCNetPVL(self, path=None):
        """Parses the control net output from hijitreg, performs smoothing,
        and sets the array of ignorable points based on smoothing and
        badness ("goodness of fit").
        """

        if path is None:
            path = self.cnet_path

        p = pvl.load(str(path))

        count = [0] * self["Columns"]

        self.control_measures = self._get_control_measures(p)

        lineCount = 0
        for i, cm in enumerate(self.control_measures):
            offset = int(i - self["BoxcarLength"] / 2)
            length = int(self["BoxcarLength"])

            if offset < 0:
                offset = 0
                length = int(self["BoxcarLength"] / 2 + i)

            if self["BoxcarLength"] > (len(self.control_measures) - i):
                # offset not changed
                length = int(
                    self["BoxcarLength"] / 2 + (len(self.control_measures) - i)
                )

            boxcar = map(
                lambda x: x["ErrorMagnitude"],
                self.control_measures[offset : offset + length],
            )

            median = statistics.median(boxcar)
            delta = abs(cm["ErrorMagnitude"] - median)

            if (
                cm["GoodnessOfFit"] > self["BadnessLimit"]
                or delta > self["ExcludeLimit"]
            ):
                self["MatchedCount"] -= 1
                self.IgnoredPoints.add(cm["PointId"])
                logger.info(
                    "Ignorable point {} with ".format(cm["PointId"])
                    + "badness {} and ".format(cm["GoodnessOfFit"])
                    + f"smoothing delta {delta}"
                )
            else:
                if "Row" in cm:
                    lineCount += 1

                if "Column" in cm:
                    count[cm["Column"]] += 1

            if len(tuple(filter(lambda x: x > 3, count))) >= 3:
                self["CanSlither"] = True
        self["MatchedLineCount"] = lineCount
        return

    @staticmethod
    def _get_control_measures(pvl) -> list:
        control_measures = list()

        # Original Perl issue: there were two "conditions" for
        # extracting information, one, labeled "<3.4" was to find
        # a ControlMeasure with a Reference = False key.  The other
        # labeled ">=3.4" was a ControlMeasure with MeasureType =
        # Candidate.  However, this condition really just ended
        # the line-by-line parsing, because the "Candidate"
        # ControlMeasure was the second one in the ControlPoint.
        # The proper logic is to get information from the
        # ControlMeasure that meets the conditions as implemented
        # below.

        for cp in pvl["ControlNetwork"].getlist("ControlPoint"):
            if "PointId" not in cp or "ControlMeasure" not in cp:
                continue
            for cm in cp.getlist("ControlMeasure"):
                if (
                    "MeasureType" not in cm
                    # or 'Reference' not in cm
                    or "GoodnessOfFit" not in cm
                    or "LineResidual" not in cm
                    or "SampleResidual" not in cm
                ):
                    continue

                if cm["MeasureType"] == "RegisteredPixel":
                    cm["ErrorMagnitude"] = math.hypot(
                        cm["SampleResidual"].value, cm["LineResidual"].value
                    )
                    # Tack on a few extra values here, and then append
                    cm["PointId"] = cp["PointId"]
                    match = re.search(
                        r"Row\s+(\d+)\s+Column\s+(\d+)", cp["PointId"]
                    )
                    if match:
                        cm["Row"] = int(match.group(1))
                        cm["Column"] = int(match.group(2))

                    control_measures.append(cm)
        return control_measures

    def filterCNetPVL(self, path=None):
        """Filters the CNET file and adds Ignored point information."""

        if len(self.IgnoredPoints) == 0:
            return

        if path is None:
            path = self.cnet_path

        p = pvl.load(str(path))

        cn = pvl.PVLModule()

        badness = 0
        for (k, v) in p["ControlNetwork"].items():
            if k == "ControlPoint":
                if (
                    v["PointId"] in self.IgnoredPoints
                    and "Ignore" not in v.keys()
                ):
                    v.append("Ignore", True)
                    badness += 1
                    logger.info("Ignoring point {}".format(v["PointId"]))
            cn.append(k, v)

        logger.info(f"{badness} point(s) ignored.")

        new_pvl = pvl.PVLModule(ControlNetwork=cn)

        with open(path, "w") as stream:
            pvl.dump(new_pvl, stream, encoder=pvl.encoder.ISISEncoder())


def HiJitReg(cube_paths: list, conf: dict, keep=False) -> list:
    cubes = list(map(hicolor.HiColorCube, cube_paths))
    red4, red5, ir10, ir11, bg12, bg13 = hicolor.separate_ccds(cubes)

    ccds = list()
    for c in red4, red5, ir10, ir11, bg12, bg13:
        if c is not None:
            ccds.append(str(c))

    logger.info(f"HiJitReg start: {', '.join(map(str, ccds))}")

    successful_ccds = list()
    if red4 is not None:
        for c in [ir10, bg12]:
            if c is not None:
                if jitter_iter(red4, c, conf, keep=keep):
                    logger.info(f"Iterations completed for {c}")
                    successful_ccds.append(c)
    if red5 is not None:
        for c in [ir11, bg13]:
            if c is not None:
                if jitter_iter(red5, c, conf, keep=keep):
                    logger.info(f"Iterations completed for {c}")
                    successful_ccds.append(c)

    # Not going to check to make sure that at most one pair fails.

    if bg12 not in successful_ccds and bg13 not in successful_ccds:
        raise RuntimeError("Registration failed for both BG halves.")

    logger.info(f"HiJitReg done: {', '.join(map(str, successful_ccds))}")
    return successful_ccds


def jitter_iter(
    red: hicolor.HiColorCube,
    color: hicolor.HiColorCube,
    conf: dict,
    keep=False,
) -> bool:
    """Iterates through hijitreg for the color cube."""

    temp_token = datetime.now().strftime("HiJitReg-%y%m%d%H%M%S")

    bin_ratio = color.bin / red.bin

    jit_param = dict()
    conf_ar = conf["AutoRegistration"]
    jit_param["GROUP"] = "HiJitReg"
    jit_param["COLS"] = conf_ar["ControlNet"]["Control_Cols"]
    jit_param["ROWS"] = conf_ar["ControlNet"]["Control_Lines"]
    jit_param["TOLERANCE"] = conf_ar["Algorithm"]["Tolerance"]
    jit_param["PATTERN_SAMPLES"] = conf_ar["PatternChip"]["Samples"]
    jit_param["PATTERN_LINES"] = conf_ar["PatternChip"]["Lines"]
    jit_param["SEARCH_SAMPLES"] = conf_ar["SearchChip"]["Samples"]
    jit_param["SEARCH_LINES"] = conf_ar["SearchChip"]["Lines"]
    jit_param["SEARCHLONGER_SAMPLES"] = conf_ar["SearchLongerChip"]["Samples"]
    jit_param["SEARCHLONGER_LINES"] = conf_ar["SearchLongerChip"]["Lines"]

    if bin_ratio > 3:
        jit_param["TOLERANCE"] -= conf_ar["Algorithm"]["INCREMENT"]

    channels = isis.getkey_k(color.path, "Instrument", "StitchedProductIds")

    coverage = 1.0

    if len(channels) < 2:
        coverage /= 2
        jit_param["COLS"] += jit_param["COLS"] / 2

    # A two-step process with completely different outcomes at each step,
    # so we can't really make a loop.
    step = 1

    logger.info(f"Attempting hijitreg iteration #{step} for {color}")

    color_jitter = JitterCube(color, conf)

    run_HiJitReg(red.path, color_jitter, jit_param, temp_token, keep=keep)

    ret = Analyze_Flat(color_jitter, step, coverage)

    if ret == -1:
        # edgy or suspect points only
        if jit_param["SEARCH_LINES"] == jit_param["SEARCHLONGER_LINES"]:
            return True
        else:
            # use larger search box for all subsequent iterations
            # (other CCDs too)
            jit_param["SEARCH_SAMPLES"] = jit_param["SEARCHLONGER_SAMPLES"]
            jit_param["SEARCH_LINES"] = jit_param["SEARCHLONGER_LINES"]
    elif ret == 0:
        # not enough points found
        # increase grid density
        jit_param["ROWS"] = jit_param["ROWS"] * 2
        if len(channels) >= 2:
            jit_param["COLS"] += 2
            coverage /= 2
    else:
        return True

    step += 1
    logger.info(f"Attempting hijitreg iteration #{step} for {color}")

    # second pass
    run_HiJitReg(red.path, color_jitter, jit_param, temp_token, keep=keep)

    # analyze output again
    ret = Analyze_Flat(color_jitter, step, coverage)

    if ret == 0:
        logger.info(f"Jitter registration failed for {color}")
        return False
    elif ret < 0:
        logger.info("!!! Validation Required !!!")
        return True
    else:
        return True


def run_HiJitReg(
    red_path: os.PathLike,
    color: JitterCube,
    params: dict,
    temptoken: str,
    keep=False,
):
    """Examine output of control net and/or flat file to automatically remove
    out-of-bound points."""

    file_status = "OVERWRITE"
    if color.regdef_path.exists():
        file_status = pvl.load(str(color.regdef_path))["AutoRegistration"][
            params["GROUP"]
        ]["File_Status"]

    if file_status == "KEEP":
        logger.info("Using existing regdef file due to KEEP file status.")
    else:
        logger.info(f"Writing new regdef file {color.regdef_path}")
        logger.info(params)
        write_regdef(color.regdef_path, params)

    tmp_control = color.cnet_path.with_suffix(".net")
    isis.hijitreg(
        red_path,
        match=color.path,
        regdef=color.regdef_path,
        rows=params["ROWS"],
        columns=params["COLS"],
        flat=color.flattab_path,
        cnet=tmp_control,
    )
    isis.cnetbin2pvl(tmp_control, to=color.cnet_path)
    if not keep:
        tmp_control.unlink()
    return


def write_regdef(out_path: os.PathLike, parameters: dict):
    """Writes PVL file that will be given to HiJitReg."""
    out_p = Path(out_path)

    pvl_text = """Object = AutoRegistration

  Version = 2

  Group = {GROUP}
     File_Status  = "OVERWRITE"
     Control_Cols = {COLS}
     Control_Rows = {ROWS}
  End_Group

  Group = Algorithm
    Name      = MaximumCorrelation
    Tolerance = {TOLERANCE}
  End_Group

  Group = PatternChip
    Samples = {PATTERN_SAMPLES}
    Lines   = {PATTERN_LINES}
  End_Group

  Group = SearchChip
    Samples = {SEARCH_SAMPLES}
    Lines   = {SEARCH_LINES}
  End_Group

End_Object

"""

    out_p.write_text(pvl_text.format(**parameters))
    return


def Analyze_Flat(
    cube: JitterCube, step: int, fraction: float, hijitreg=True
) -> int:
    cube.reset()
    cube.filterCNetPVL()

    logger.info(
        "Matched Registers     = {} of {}".format(
            cube["MatchedCount"], cube["RegisterCount"]
        )
    )
    logger.info("Average Sample Offset = {}".format(cube["AvgSampleOffset"]))
    logger.info("Average Line Offset   = {}".format(cube["AvgLineOffset"]))
    logger.info("Edgy Count            = {}".format(cube["EdgyCount"]))
    logger.info("Suspect Points        = {}".format(cube["SuspectCount"]))

    if cube["AvgSampleOffset"] is None or cube["AvgLineOffset"] is None:
        logger.warning("No points met the correlation tolerance.")
        return 0

    if hijitreg and cube["CanSlither"] is False:
        logger.warning(
            "Too few correlated lines found for cubic slither fit."
        )
        return 0

    good_fraction = (cube["MatchedCount"] - cube["SuspectCount"]) / cube[
        "RegisterCount"
    ]

    if good_fraction < 0.5 * fraction and step <= 1:
        logger.info(
            f"Too few correlated points ({good_fraction}) "
            "found at this tolerance."
        )
        return 0

    elif hijitreg and good_fraction < 0.25 * fraction and step <= 2:
        logger.info(
            f"Too few correlated points ({good_fraction}) "
            "found at this tolerance."
        )
        return -1

    if hijitreg and cube["EdgyCount"] > 2 and good_fraction > 0.8 * fraction:
        logger.info("More than two edgy points with search box size.")
        return -1

    if cube["SuspectCount"] > 3:
        logger.info("More than three suspect points with search box size.")
        return -1

    return 1



#!/usr/bin/env python
"""Normalizes HiRISE Color products.

This program will normalize the BG and IR color across the left and right
halves of each set of color CCDs. It uses the HiSlither COLOR4 and COLOR5 cubes.


Data Flow
---------
Input Products:

- ``COLOR4`` and ``COLOR5`` files which are the result of HiSlither.

Output Products:

- creates UNFILTERED versions of the COLOR4 and COLOR5 files.
- creates HiColorNorm versions of the COLOR4 and COLOR5 files.

"""

# Copyright 2006-2020, Arizona Board of Regents on behalf of the Lunar and
# Planetary Laboratory at the University of Arizona.
#   - Orignal Perl program.
#
# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#   - Elements of this Python program are are based on the original Perl
#     but the logic here is rewritten from scratch to emulate functionality.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This program is based on HiColorNorm version 1.6.2 (2020/04/28),
# and on the Perl HiColorNorm program ($Revision: 1.28 $
#                                      $Date: 2020/04/28 16:17:31 $)
# by Eric Eliason as an employee of the University of Arizona.

import argparse
import collections
import csv
import json
import logging
import os
import pkg_resources
import re
import shutil
import statistics
from datetime import datetime
from itertools import repeat
from pathlib import Path

import pvl

import kalasiris as isis
import hiproc.hirise as hirise
import hiproc.util as util

logger = logging.getLogger(__name__)


class ColorCube(hirise.ObservationID):
    """A class for HiRISE multiband COLOR cubes."""

    def __init__(self, pathlike, dbs=None):
        self.path = Path(pathlike)
        match = re.search(r"COLOR(\d)", str(pathlike))
        if match:
            self.ccdnumber = match.group(1)
        else:
            raise ValueError(f"Could not extract a COLOR from {pathlike}")

        super().__init__(hirise.get_ObsID_fromfile(self.path))
        self.lines = int(isis.getkey_k(self.path, "Dimensions", "Lines"))
        self.samps = int(isis.getkey_k(self.path, "Dimensions", "Samples"))
        self.bands = int(isis.getkey_k(self.path, "Dimensions", "Bands"))

        if self.bands != 3:
            raise Exception("{} must have 3 bands.".format(self.path))

        # centers = isis.getkey_k(self.path, 'BandBin', 'Center')
        # match_c = re.search(r'900,\s+700,\s+500', centers)
        if not match:
            raise Exception(
                f"{self.path} must have BandBin/Center of 900, 700, 500."
            )
        self.band = {"IR": 1, "RED": 2, "BG": 3}

        self.pids = isis.getkey_k(
            self.path, "Mosaic", "SourceProductId"
        ).split(", ")
        self.ir_bin = self.get_binning("IR", dbs)
        self.red_bin = self.get_binning("RED", dbs)
        # self.bg_bin = get_binning('BG')

        self.mask_path = {"IR": None, "BG": None}
        self.crop_path = {"IR": None, "BG": None}
        self.nrm_path = {"IR": None, "BG": None}

        self.crop_sline = None
        self.crop_lines = None

        self.final_path = None

    def __str__(self):
        return "_".join([super().__str__(), self.get_ccd()])

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.path}')"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return super().__eq__(other) and self.ccdnumber == other.ccdnumber
        return False

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            if super().__eq__(other):
                return int(self.ccdnumber) < int(other.ccdnumber)
            else:
                return super().__lt__(other)
        else:
            return NotImplemented

    def get_ccd(self) -> str:
        return f"COLOR{self.ccdnumber}"

    def get_obsid(self) -> hirise.ObservationID:
        return hirise.ObservationID(
            self.phase, self.orbit_number, self.target
        )

    def set_crop_lines(self, conf):
        sline = conf["HiColorNorm"]["HiColorNorm_Crop_Top"] + 1
        nline = (
            self.lines
            - conf["HiColorNorm"]["HiColorNorm_Crop_Top"]
            - conf["HiColorNorm"]["HiColorNorm_Crop_Bot"]
        )
        if nline < 5000:
            sline = 1
            nline = self.lines

        self.crop_sline = sline
        self.crop_lines = nline
        return (sline, nline)

    def get_binning_from(self, color_code, pids=None, db_paths=None):
        dbs = list()
        if pids is None:
            pids = self.pids

        parent = self.path.parent
        if db_paths is None:
            db_paths = list()
            for pid in pids:
                db_paths.extend(list(parent.glob(f"{pid}*.json")))

        for pid in pids:
            if color_code in pid:
                if len(db_paths) > 0:
                    for p in db_paths:
                        with open(p, "r") as f:
                            dbs.append(json.load(f))
                else:
                    for p in parent.glob(f"{pid}*.cub"):
                        temp_d = dict()
                        temp_d["PRODUCT_ID"] = isis.getkey_k(
                            p, "Archive", "ProductId"
                        )
                        temp_d["BINNING"] = int(
                            isis.getkey_k(p, "Instrument", "Summing")
                        )
                        dbs.append(temp_d)
        return dbs

    def get_binning(self, color_code, dbs=None, pids=None, db_paths=None):
        # Get the binning values for the component red, ir, and bg
        # observations.
        if pids is None:
            pids = self.pids

        if dbs is None:
            dbs = self.get_binning_from(color_code, pids, db_paths)

        bins = collections.Counter()
        for db in dbs:
            if (
                "PRODUCT_ID" in db
                and db["PRODUCT_ID"] in pids
                and color_code in db["PRODUCT_ID"]
            ):
                bins[db["BINNING"]] += 1

        if len(bins) != 1:
            raise Exception(
                "Gathered {} different ".format(len(bins))
                + f"binning values for the {color_code} channel, "
                "must be an error."
            )
        return int([*bins][0])

    def get_boxcar_size(self, color_bin: int) -> int:
        if color_bin > self.red_bin * 2:
            return 5
        elif color_bin > self.red_bin:
            return 3
        else:
            return 1


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        parents=[util.parent_parser()],
        conflict_handler="resolve",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "-o", "--output",
        required=False,
        default="_COLOR.cub",
        help="The filename to be used for the output color cube.  If it "
             "begins with an underscore ('_') it will be assumed to be a "
             "suffix that will be appended to a name derived from the "
             "observation. Default: %(default)s"
    )
    parser.add_argument(
        "-c",
        "--conf",
        required=False,
        type=argparse.FileType('r'),
        default=pkg_resources.resource_stream(
            __name__,
            'data/HiColorNorm.conf'
        ),
        help="Path to the HiColorNorm config file.  Defaults to "
             "HiColorNorm.conf distributed with the library."
    )
    parser.add_argument(
        "-n",
        "--nounfiltered",
        action="store_false",
        dest="Make_Unfiltered",
        help="Stops creation of an unfiltered cube.",
    )
    parser.add_argument(
        "cubes",
        type=Path,
        nargs=2,
        metavar="COLOR-cube",
        help="The COLOR4.cub and COLOR5.cub files"
    )
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    with util.main_exceptions(args.verbose):
        (ir_ratio, bg_ratio) = HiColorNorm(
            args.cubes,
            args.output,
            pvl.load(args.conf),
            make_unfiltered=args.Make_Unfiltered,
            keep=args.keep,
        )

    # Original Perl connects to HiCat to query the binning for CCDs
    #   and also whether there were furrows in the IR or BG components
    #   Finally, there's an insert into the Color_Processing_Statistics
    #       table that puts in the
    #       HICOLORNORM_RATIO_CORRECTION_STANDARD_DEVIATION for each FILTER
    # At the moment, we're not doing that here.


def conf_check(conf: dict) -> None:
    """Various checks on parameters in the configuration."""

    t = "HiColorNorm_NoiseFilter_IR10"
    util.conf_check_bool(t, conf["HiColorNorm"][t])

    t = "HiColorNorm_Make_Stitch"
    util.conf_check_bool(t, conf["HiColorNorm"][t])

    # This is not set via conf files, but via command line parameter.
    # t = 'HiColorNorm_Make_Unfiltered_Cube'
    # util.conf_check_bool(t, conf['HiColorNorm'][t])

    t = "HiColorNorm_Normalization_Minimum"
    util.conf_check_bounds(t, (-16384.0, 16364.0), conf["HiColorNorm"][t])

    t = "HiColorNorm_Normalization_Maximum"
    util.conf_check_bounds(t, (-16384.0, 16364.0), conf["HiColorNorm"][t])

    return


def set_outpath(out: os.PathLike, cubes) -> Path:
    # Check that they all have the same Obs ID:
    if len(set(c.get_obsid() for c in cubes)) != 1:
        raise ValueError(
            "These cube files don't all have the same "
            f"Observation ID: {cubes}"
        )

    if str(out).startswith("_"):
        return cubes[0].path.parent / Path(str(cubes[0].get_obsid()) + out)
    else:
        return Path(out)


def HiColorNorm(
    cubes: list,
    output,
    conf: dict,
    make_unfiltered=True,
    db_list=None,
    keep=False
):
    logger.info(f"HiColorNorm start: {cubes}")

    # GetConfigurationParameters()
    conf_check(conf)

    cubes = list(map(ColorCube, cubes, repeat(db_list)))
    cubes.sort()

    outcub_path = set_outpath(output, cubes)

    temp_token = datetime.now().strftime("HiColorNorm-%y%m%d%H%M%S")
    out_p = Path(outcub_path)

    furrow_flag = FurrowCheck(cubes)

    to_del = isis.PathSet()

    for i, _ in enumerate(cubes):
        cubes[i].set_crop_lines(conf)

    for i, c in enumerate(cubes):
        # Protect the processing from erroneous/spurious pixels
        mask_list = list()
        for b in (1, 2, 3):
            tmp_p = to_del.add(
                c.path.with_suffix(f".{temp_token}.temp{b}.cub")
            )
            isis.mask(
                f"{c.path}+{b}",
                mask=f"{c.path}+{b}",
                to=tmp_p,
                minimum=0.0,
                maximum=2.0,
                preserve="INSIDE",
            )
            mask_list.append(tmp_p)

        c.final_path = c.path.with_suffix(".HiColorNorm.cub")
        isis.cubeit_k(mask_list, to=c.final_path)

        (cubes[i].mask_path["IR"], cubes[i].crop_path["IR"]) = per_color(
            c, temp_token, "IR", keep=keep
        )
        (cubes[i].mask_path["BG"], cubes[i].crop_path["BG"]) = per_color(
            c, temp_token, "BG", keep=keep
        )

    ir_ratio_stddev = per_band(
        cubes, out_p, temp_token, "IR", furrow_flag, make_unfiltered, keep=keep
    )
    bg_ratio_stddev = per_band(
        cubes, out_p, temp_token, "BG", furrow_flag, make_unfiltered, keep=keep
    )

    if conf["HiColorNorm"]["HiColorNorm_Make_Stitch"]:
        # listpath = to_del.add(c.path.with_suffix(f'.{temp_token}.list.txt'))
        # listpath.write_text(
        #   '\n'.join(str(c.final_path) for c in cubes) + '\n')

        with isis.fromlist.temp([str(c.final_path) for c in cubes]) as f:
            isis.hiccdstitch(fromlist=f, to=out_p)

        for c in cubes:
            to_del.add(c.final_path)

    if not keep:
        to_del.unlink()
        for c in cubes:
            for cc in ("IR", "BG"):
                c.mask_path[cc].unlink()
                c.crop_path[cc].unlink()
                c.nrm_path[cc].unlink()

    logger.info("HiColorNorm done.")
    return ir_ratio_stddev, bg_ratio_stddev


def FurrowCheck(cubes, db_path=None) -> bool:
    # If any of these cubes have been flagged as having furrows
    if db_path is None:
        db_path = list()
        for c in cubes:
            parent = c.path.parent
            for pid in c.pids:
                if "RED" in pid:
                    continue
                for p in parent.glob(f"{pid}*.json"):
                    db_path.append(p)
    dbs = list()
    for p in db_path:
        with open(p) as f:
            dbs.append(json.load(f))

    furrow = any(bool(x.get("zapped", False)) for x in dbs)
    return furrow


def per_color(cube, temp_token, color_code, keep=False):

    tt = f"{temp_token}_{color_code}"

    numerator = "{}+{}".format(cube.path, cube.band[color_code])
    denominator = "{}+{}".format(cube.path, cube.band["RED"])

    # Generate the IR/RED and BG/RED ratios for each of the COLOR products
    rat_p = cube.path.with_suffix(f".{tt}.ratio.cub")
    isis.ratio(num=numerator, den=denominator, to=rat_p)

    # mask out invalid pixels
    # Possible future update: Make mosaic of ratio files then run cubnorm
    # correction on these for the unfiltered products, to avoid any null
    # pixels created during the ratio process.
    mask_p = cube.path.with_suffix(f".{tt}.mask.cub")
    isis.mask(
        rat_p,
        mask=rat_p,
        to=mask_p,
        minimum=0.0,
        maximum=4.0,
        preserve="INSIDE",
    )

    # Generate the crop files
    crop_p = cube.path.with_suffix(f".{tt}.ratcrop.cub")
    isis.crop(mask_p, to=crop_p, line=cube.crop_sline, nlines=cube.crop_lines)

    if not keep:
        rat_p.unlink()

    return mask_p, crop_p


def make_LR_mosaic(
    left_path, right_path, left_samps, mosaic_path, lines, samps
):
    isis.handmos(
        left_path,
        mosaic=mosaic_path,
        create="YES",
        nlines=lines,
        nsamp=samps,
        nbands=1,
        outsamp=1,
        outline=1,
        outband=1,
    )
    isis.handmos(
        right_path,
        mosaic=mosaic_path,
        outsamp=(left_samps + 1),
        outline=1,
        outband=1,
    )
    return


def cubenorm_stats(crpmos, mos, mosnrm, keep=False):
    # get cubenorm statistics on the crop mosaic files
    stats_p = Path(crpmos).with_suffix(".cubenorm.txt")
    isis.cubenorm(
        crpmos,
        stats=stats_p,
        format="table",
        direction="column",
        norm="average",
    )
    averages = list()
    with open(stats_p) as csvfile:
        reader = csv.DictReader(csvfile, dialect=isis.cubenormfile.Dialect)
        for row in reader:
            averages.append(float(row["Average"]))
    stddev = statistics.stdev(averages)

    # Now apply the statistics to the non-crop image
    isis.cubenorm(
        mos,
        fromstats=stats_p,
        to=mosnrm,
        preserve=True,
        statsource="table",
        format="table",
        direction="column",
        norm="average",
    )

    if not keep:
        stats_p.unlink()

    return stddev


def make_unfiltered(
    in_path, nrm_path, temp_token, color_code, color_band, keep=False
):
    tt = f"{temp_token}_{color_code}"
    in_p = Path(in_path)
    unfiltered_p = in_p.with_name(
        in_p.name.replace("COLOR", "UNFILTERED_COLOR")
    )
    shutil.copyfile(in_p, unfiltered_p)
    # run ISIS algebra program to create unfiltered
    # (normalized-IR/RED)*RED files
    alg_unfiltered_path = in_p.with_suffix(f".{tt}.algebra.cub")
    isis.algebra(
        nrm_path,
        from2="{}+2".format(in_p),
        to=alg_unfiltered_path,
        operator="MULTIPLY",
    )
    # Update the output file with the non-filtered normalized IR and BG bands
    isis.handmos(
        alg_unfiltered_path,
        mosaic=unfiltered_p,
        outsample=1,
        outline=1,
        matchbandbin=False,
        outband=color_band,
    )
    if not keep:
        alg_unfiltered_path.unlink()
    return unfiltered_p


def lpfz_triplefilter(
    from_path: os.PathLike, to_path: os.PathLike, keep=False
) -> None:
    to_del = isis.PathSet()
    from_p = Path(from_path)
    z1 = to_del.add(from_p.with_suffix(".z1.cub"))
    lpfz_filtering(from_p, z1, 11, 5)

    z2 = to_del.add(from_p.with_suffix(".z2.cub"))
    lpfz_filtering(z1, z2, 21, 9)

    lpfz_filtering(z2, to_path, 41, 11)

    if not keep:
        to_del.unlink()

    return


def lpfz_filtering(
    from_path: os.PathLike, to_path: os.PathLike, boxl: int, boxs: int
) -> None:
    isis.lowpass(
        from_path,
        to=to_path,
        lines=boxl,
        samples=boxs,
        filter="OUTSIDE",
        minopt="PERCENTAGE",
        minimum=25,
        low=0.00,
        high=2.0,
        null=True,
        HRS=True,
        HIS=True,
        LIS=True,
        LRS=True,
    )


def per_band(
    cubes, out_p, temp_token, color_code, furrow_flag, unfiltered, keep=False
) -> float:
    to_del = isis.PathSet()

    # Generate the handmos of both the ratio and the croped ratio files
    if len(cubes) == 2:
        maxlines = max(c.lines for c in cubes)
        maxsamps = sum(c.samps for c in cubes)

        # Generate the mosaic of the ratio image
        mos_p = to_del.add(
            out_p.with_suffix(f".{temp_token}.{color_code}.mos.cub")
        )
        make_LR_mosaic(
            cubes[0].mask_path[color_code],
            cubes[1].mask_path[color_code],
            cubes[0].samps,
            mos_p,
            maxlines,
            maxsamps,
        )

        # Generate the mosaic of the crop ratio image
        maxlines = max(c.crop_lines for c in cubes)
        crpmos_p = to_del.add(
            out_p.with_suffix(f".{temp_token}.{color_code}.moscrop.cub")
        )
        make_LR_mosaic(
            cubes[0].crop_path[color_code],
            cubes[1].crop_path[color_code],
            cubes[0].samps,
            crpmos_p,
            maxlines,
            maxsamps,
        )
    else:
        mos_p = cubes[0].mask_path[color_code]
        crpmos_p = cubes[0].crop_path[color_code]

    mosnrm_p = to_del.add(
        out_p.with_suffix(f".{temp_token}.{color_code}.mosnorm.cub")
    )
    ratio_stddev = cubenorm_stats(crpmos_p, mos_p, mosnrm_p, keep=keep)

    # from the handmos cubes, pull out the individual CCDs with crop
    if len(cubes) == 2:
        samp = 1
        for i, c in enumerate(cubes):
            cubes[i].nrm_path[color_code] = c.path.with_suffix(
                f".{temp_token}.{color_code}.norm.cub"
            )
            isis.crop(
                mosnrm_p,
                to=cubes[i].nrm_path[color_code],
                sample=samp,
                nsamples=c.samps,
            )
            samp += c.samps
    else:
        # If there was only one file then all we need to do is rename files
        cubes[0].nrm_path[color_code] = mosnrm_p

    if unfiltered:
        # Create the unfiltered color cubes, if needed
        for c in cubes:
            make_unfiltered(
                c.path,
                c.nrm_path[color_code],
                temp_token,
                color_code,
                c.band[color_code],
                keep=keep,
            )

    # low pass filter the files
    for c in cubes:
        # If $lpfz=1 then we're only interpolating null pixels caused by
        # noise or furrow removal

        lowpass_args = dict(
            minopt="PERCENT",
            minimum=25,
            low=0.00,
            high=2.0,
            null=True,
            hrs=True,
            lis=True,
            lrs=True,
        )

        # As the comment below states, I *think* that the OP had a copy
        # and paste error here, because even though it kept track of both
        # the IR and BG boxcar dimensions, they were always identical.
        logger.warning(
            "The original Perl calculates the boxcar size for "
            "both the IR and BG based only on the ratio of the "
            "IR to the RED."
        )
        if c.get_boxcar_size(c.ir_bin) == 1:
            lowpass_args["lines"] = 3
            lowpass_args["samples"] = 3
            lowpass_args["filter"] = "OUTSIDE"
        else:
            lowpass_args["lines"] = 1
            lowpass_args["samples"] = 1

        lpf_path = to_del.add(
            c.path.with_suffix(f".{temp_token}.{color_code}.lpf.cub")
        )
        isis.lowpass(c.nrm_path[color_code], to=lpf_path, **lowpass_args)

        # Perform lpfz filters to interpolate null pixels due to furrows or
        #   bad pixels
        if furrow_flag:
            lpfz_path = c.path.with_suffix(
                f".{temp_token}.{color_code}.lpfz.cub"
            )
            lpfz_triplefilter(lpf_path, lpfz_path, temp_token, keep=keep)
        else:
            lpfz_path = lpf_path

        # run ISIS algebra program to created (normalized-IR/RED)*RED files
        alg_path = to_del.add(
            c.path.with_suffix(f".{temp_token}.{color_code}.algebra.cub")
        )
        isis.algebra(
            lpfz_path, from2=f"{c.path}+2", to=alg_path, operator="MULTIPLY"
        )

        # Update the output file with the normalized IR and BG bands
        isis.handmos(
            alg_path,
            mosaic=c.final_path,
            outsample=1,
            outline=1,
            outband=c.band[color_code],
            matchbandbin=False,
        )

    if not keep:
        to_del.unlink()

    return ratio_stddev


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""This module has tests for the HiRISE HiccdStitch functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import contextlib
import pkg_resources
import subprocess
import unittest
from pathlib import Path
from unittest.mock import call
from unittest.mock import mock_open
from unittest.mock import patch

import pvl

# import hiproc.hirise as hirise
import hiproc.HiccdStitch as hcs

conf = pvl.load(
    pkg_resources.resource_stream(
        "hiproc",
        "data/HiccdStitch.conf"
    )
)


def getkey(cube, group, key):
    values = {
        "ProductId": None,
        "Summing": 2,
        "Lines": 1024,
        "Samples": 1024,
        "Special_Processing_Flag": "NOMINAL",
    }
    return values[key]


class TestHiccdStitchCube(unittest.TestCase):
    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_init(self, mock_getkey):
        c = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5_0")
        self.assertTrue(c.ns, 1024)

    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_set_cubenorm_lines(self, mock_getkey):

        c = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5_0")
        self.assertTrue(
            (1.5, 1023.0), c.set_cubenorm_lines(1, 1, None, None, 1)
        )
        self.assertRaises(
            ValueError, c.set_cubenorm_lines, 1, 1, 2000, None, 2
        )
        self.assertRaises(
            ValueError, c.set_cubenorm_lines, 1, 1, None, 2000, 2
        )
        self.assertRaises(
            RuntimeError, c.set_cubenorm_lines, 1, 1, 1000, 800, 2
        )

    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_set_balance(self, mock_getkey):
        c = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5_0")
        c.set_balance(1, 1, {2: [1, 1]}, 1)
        self.assertEqual(2, c.ss_balance_left)
        self.assertEqual(1, c.ns_balance_left)
        self.assertEqual(1023, c.ss_balance_right)
        self.assertEqual(1, c.ns_balance_right)
        self.assertEqual(1, c.sl_balance)
        self.assertEqual(0, c.nl_balance)

    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_set_ls_path(self, mock_getkey):
        c = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5_0")
        p = "dummy-path"
        c.set_ls_path(p)
        self.assertEqual(p, c.ls_path)
        self.assertEqual(p, c.lm_path)

    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_set_rs_path(self, mock_getkey):
        c = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5_0")
        p = "dummy-path"
        c.set_rs_path(p)
        self.assertEqual(p, c.rs_path)
        self.assertEqual(p, c.rm_path)

    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_gather_from_db(self, mock_getkey):
        c = hcs.HiccdStitchCube("tmp/PSP_010502_2090_RED5_0")
        db = {"hical_status": "BadCal", "IMAGE_SIGNAL_TO_NOISE_RATIO": 5}
        with patch("hiproc.HiccdStitch.json.load", return_value=db):
            with patch(
                "hiproc.HiccdStitch.Path.glob", return_value=["d", "d"]
            ):
                with patch("hiproc.HiccdStitch.open", mock_open()):
                    c.gather_from_db()
                    self.assertEqual("BadCal", c.hical_status)
                    self.assertListEqual([5.0, 5.0], c.snr_list)


class TestConf(unittest.TestCase):
    def test_conf_check(self):
        self.assertIsNone(hcs.conf_check(conf))

    def test_make_area_dict(self):
        # print(hcs.make_area_dict(c['HiccdStitch']))
        truth = {
            1: [12, 36],
            2: [6, 18],
            3: [5, 15],
            4: [5, 9],
            8: [5, 5],
            16: [1, 2],
        }
        self.assertDictEqual(truth, hcs.make_area_dict(conf["HiccdStitch"]))


class TestBasic(unittest.TestCase):
    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_set_outpath(self, mock_getkey):
        c1 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5_0")
        c2 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED3_1")
        diff = hcs.HiccdStitchCube("tmp/PSP_005632_1225_RED5_1")

        self.assertRaises(ValueError, hcs.set_outpath, "dummy", [c1, c2, diff])

        out = Path("foo")
        self.assertEqual(out, hcs.set_outpath(out, [c1, c2]))

        truth = Path("dummy/PSP_010502_2090_RED.foo")
        self.assertEqual(truth, hcs.set_outpath(".foo", [c1, c2]))

    @patch("hiproc.HiStitch.isis.enlarge")
    @patch("hiproc.HiStitch.isis.crop")
    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_crop_and_scale(self, mock_getkey, mock_crop, moc_enlarge):
        c1 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5_0")
        c2 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED3_1")
        c1.smag = 1
        c1.lmag = 1
        c2.smag = 1
        c2.lmag = 1
        new_c, to_del = hcs.crop_and_scale([c1, c2])
        p = Path("dummy/PSP_010502_2090_RED5_0.left.crop.cub")
        self.assertEqual(p, new_c[0].ls_path)
        self.assertIn(p, to_del)

    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_get_group_i(self, mock_getkey):
        c3 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED3")
        c5 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5")
        c6 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED6")
        c3.rstats = 1
        c3.lstats = 1
        c5.rstats = 1
        c5.lstats = 1
        c6.lstats = 1
        c6.rstats = 1
        pairs = hcs.get_group_i([c5, c3, c6])
        self.assertListEqual([3], pairs[0][1])
        self.assertListEqual([5, 6], pairs[1][1])

    # Not testing hcs.get_stats(), because it just straight up calls ISIS
    # routines.

    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_get_correction(self, mock_getkey):
        c5 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5")
        c6 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED6")
        c5.correction = 2
        c5.rstats = 2
        c6.lstats = 2
        self.assertEqual(2, hcs.get_correction(c6, c5, "MULTIPLY", 6))
        self.assertEqual(1, hcs.get_correction(c6, c5, "MULTIPLY", 0))
        self.assertEqual(0, hcs.get_correction(c6, c5, "ADD", 0))

    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_get_normalization(self, mock_getkey):
        c1 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED1")
        c2 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED2")
        c3 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED3")
        c4 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED4")
        c5 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5")
        c6 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED6")
        c1.correction = 1
        c2.correction = 2
        c3.correction = 3
        c4.correction = 4
        c5.correction = 5
        c6.correction = 6
        self.assertEqual(
            5,
            hcs.get_normalization(
                [c1, c2, c3, c4, c5, c6], [1, 2, 3, 4, 5, 6], -1, [5, 11, 13]
            ),
        )
        self.assertEqual(
            3, hcs.get_normalization([c2, c3, c4], [2, 3, 4], -2, [5, 11, 13])
        )


class TestMock(unittest.TestCase):
    def setUp(self):
        self.cubenorm_data = """    Band  RowCol    ValidPoints        Average         Median         StdDev        Minimum        Maximum
       1       1           4000       0.154037       0.154375     0.00549645       0.132305       0.173675
       1       2           4000       0.151859       0.152177     0.00559696       0.127681       0.172393"""
        hiconf = {
            "HiccdStitch_Bin1_Skip_Top_Lines": 1,
            "HiccdStitch_Bin1_Skip_Bot_Lines": 1,
            "HiccdStitch_Bin01_Area": [1, 1],
            "HiccdStitch_Bin02_Area": [1, 1],
            "HiccdStitch_Bin03_Area": [1, 1],
            "HiccdStitch_Bin04_Area": [1, 1],
            "HiccdStitch_Bin08_Area": [1, 1],
            "HiccdStitch_Bin16_Area": [1, 1],
            "HiccdStitch_Balance": True,
            "HiccdStitch_Balance_Correction": "MULTIPLY",
            "HiccdStitch_Cubenorm_Method": "DIVIDE",
            "HiccdStitch_Interpolation": "BILINEAR",
            "HiccdStitch_Normalization_Minimum": 0.0,
            "HiccdStitch_Normalization_Maximum": 1.5,
            "HiccdStitch_Control_CCD": [5, 11, 13],
            "HiccdStitch_SNR_Threshold": 50,
            "HiccdStitch_Clean": "KEEP",
            "HiccdStitch_Version_Enable": False,
            "HiccdStitch_Balance_Method": "AVERAGE"
        }
        self.conf = {"HiccdStitch": hiconf}

    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_GetImageDims(self, mock_getkey):
        c1 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5_0")
        c2 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED3_1")
        cubes = [c1, c2]
        new_c = hcs.GetImageDims(cubes, self.conf, None, None)
        self.assertEqual(1, new_c[0].smag)
        self.assertEqual(1, new_c[0].lmag)

    def test_AnalyzeStats(self):
        m = mock_open(read_data=self.cubenorm_data)
        m.return_value.__iter__ = lambda self: self
        m.return_value.__next__ = lambda self: next(iter(self.readline, ""))
        with patch("hiproc.HiccdStitch.open", m):
            self.assertAlmostEqual(
                2.371842000000029e-06,
                hcs.AnalyzeStats("dummy_cubenormout", "dummy_filtered"),
                places=10,
            )

    @patch("hiproc.HiStitch.isis.PathSet.unlink")
    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    @patch("hiproc.HiStitch.isis.cubenorm")
    @patch("hiproc.HiStitch.isis.crop")
    def test_CubeNormStep(
        self, mock_crop, mock_cubenorm, mock_getkey, mock_PathSetunlink
    ):
        m = mock_open(read_data=self.cubenorm_data)
        m.return_value.__iter__ = lambda self: self
        m.return_value.__next__ = lambda self: next(iter(self.readline, ""))
        with patch("hiproc.HiccdStitch.open", m):
            c = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5_0")
            new_c = hcs.CubeNormStep(c, self.conf["HiccdStitch"])
            self.assertEqual(
                Path("dummy/PSP_010502_2090_RED5_0.cubenorm.cub"),
                new_c.nextpath,
            )

    @patch("hiproc.HiStitch.isis.algebra")
    @patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey)
    def test_make_balance(self, mock_getkey, mock_algebra):
        c = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5_0")
        c.correction = 1
        hcs.make_balance(
            c, self.conf["HiccdStitch"], Path("dummy_balance.cub")
        )
        mock_algebra.assert_has_calls(
            [
                call(
                    Path("dummy/PSP_010502_2090_RED5_0"),
                    a=1,
                    c=0,
                    operator="unary",
                    to="dummy_balance.cub+SignedWord+0.0:1.5",
                )
            ]
        )

        mock_algebra.reset_mock()
        new_conf = self.conf["HiccdStitch"]
        new_conf["HiccdStitch_Balance_Correction"] = "ADD"
        hcs.make_balance(c, new_conf, Path("dummy_balance.cub"))
        mock_algebra.assert_has_calls(
            [
                call(
                    Path("dummy/PSP_010502_2090_RED5_0"),
                    a=1,
                    c=1,
                    operator="unary",
                    to="dummy_balance.cub+SignedWord+0.0:1.5",
                )
            ]
        )

    @patch(
        "hiproc.HiccdStitch.pvl.loads",
        return_value={"Results": {"Average": 1}},
    )
    @patch("hiproc.HiStitch.isis.PathSet.unlink")
    @patch("hiproc.HiccdStitch.isis.stats")
    @patch("hiproc.HiccdStitch.isis.enlarge")
    @patch("hiproc.HiccdStitch.isis.crop")
    @patch("hiproc.HiccdStitch.isis.algebra")
    @patch("hiproc.HiccdStitch.isis.mask")
    @patch("hiproc.HiccdStitch.isis.getkey_k", side_effect=getkey)
    def test_BalanceStep(
        self,
        m_getkey,
        m_mask,
        m_algebra,
        m_crop,
        m_enlarge,
        m_stats,
        m_unlink,
        m_loads,
    ):
        c1 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED1")
        c2 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED2")
        c3 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED3")
        c4 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED4")
        c5 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED5")
        c6 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED6")

        cubes = hcs.BalanceStep(
            [c1, c2, c3, c4, c5, c6], self.conf["HiccdStitch"]
        )
        self.assertEqual(1, cubes[0].lstats)
        self.assertEqual(1, cubes[0].correction)
        self.assertEqual(
            Path("dummy/PSP_010502_2090_RED1.balance.cub"), cubes[0].nextpath
        )

    @patch("hiproc.HiccdStitch.isis.editlab")
    @patch("hiproc.HiccdStitch.isis.getkey_k", side_effect=getkey)
    def test_SpecialProcessingFlags(self, m_getkey, m_editlab):
        p = Path("dummy/PSP_010502_2090_RED1")
        c = hcs.HiccdStitchCube(p)
        hcs.SpecialProcessingFlags(c)
        m_editlab.assert_has_calls(
            [
                call(
                    p,
                    grpname="Instrument",
                    keyword="Special_Processing_Flag",
                    option="MODKEY",
                    value="NOMINAL",
                )
            ]
        )
        m_editlab.reset_mock()
        c.hical_status = "BadCal"
        hcs.SpecialProcessingFlags(c)
        m_editlab.assert_has_calls(
            [
                call(
                    p,
                    grpname="Instrument",
                    keyword="Special_Processing_Flag",
                    option="MODKEY",
                    value="BADCAL",
                )
            ]
        )
        m_editlab.reset_mock()
        c.cubenormstep = True
        hcs.SpecialProcessingFlags(c)
        m_editlab.assert_has_calls(
            [
                call(
                    p,
                    grpname="Instrument",
                    keyword="Special_Processing_Flag",
                    option="MODKEY",
                    value="CUBENORM",
                )
            ]
        )
        with patch(
            "hiproc.HiccdStitch.isis.getkey_k",
            side_effect=subprocess.CalledProcessError(cmd="foo", returncode=1),
        ):
            m_editlab.reset_mock()
            hcs.SpecialProcessingFlags(c)
            m_editlab.assert_has_calls(
                [
                    call(
                        p,
                        grpname="Instrument",
                        keyword="Special_Processing_Flag",
                        option="ADDKEY",
                        value="CUBENORM",
                    )
                ]
            )

    @patch("hiproc.HiccdStitch.isis.getkey_k", side_effect=getkey)
    def test_SNR_Check(self, m_getkey):
        c1 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED1")
        c2 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED2")
        self.assertIsNone(hcs.SNR_Check([c1, c2], 50))
        c1.bin = c2.bin = 1
        c1.snr_list = [5]
        c2.snr_list = [3, 4]
        with self.assertLogs(level="WARN"):
            self.assertIsNone(hcs.SNR_Check([c1, c2], 50))

    @patch("hiproc.HiccdStitch.SNR_Check")
    @patch("hiproc.HiccdStitch.isis.hiccdstitch")
    @patch("hiproc.HiccdStitch.SpecialProcessingFlags")
    @patch(
        "hiproc.HiccdStitch.BalanceStep",
        side_effect=lambda a, b, keep=False: a,
    )
    @patch("hiproc.HiccdStitch.CubeNormStep", side_effect=lambda a, b, c: a)
    @patch("hiproc.HiStitch.isis.PathSet.unlink")
    @patch("hiproc.HiccdStitch.isis.getkey_k", side_effect=getkey)
    def test_HiccdStitch(
        self,
        m_getkey,
        m_unlink,
        m_CubeNormStep,
        m_Balance,
        m_Special,
        m_hiccdstitch,
        m_SNR,
    ):
        c1 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED1")
        c2 = hcs.HiccdStitchCube("dummy/PSP_010502_2090_RED2")
        p = "out-dummy/foo.cub"
        hcs.HiccdStitch([c1, c2], p, self.conf)
        m_CubeNormStep.assert_not_called()
        m_Balance.assert_has_calls(
            [call([c1, c2], self.conf["HiccdStitch"], keep=False)]
        )
        m_Special.assert_has_calls([call(c1), call(c2)])
        m_hiccdstitch.assert_called_once()
        m_SNR.assert_has_calls([call([c1, c2], 50)])



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

requirements = [
    "gdal>=3.0.1",
    "kalasiris>=1.8.0",
    "matplotlib>=3.2.1",
    "multidict",
    "numpy>=1.18.1",
    "pvl>=1.0.1",
    "scipy>=1.4.1"
]

setup(
    entry_points={
        'console_scripts': [
            'EDR_Stats=hiproc.EDR_Stats:main',
            'HiBeautify=hiproc.HiBeautify:main',
            'HiCal=hiproc.HiCal:main',
            'HiColorInit=hiproc.HiColorInit:main',
            'HiColorNorm=hiproc.HiColorNorm:main',
            'HiJACK=hiproc.HiJACK:main',
            'HiJitReg=hiproc.HiJitReg:main',
            'HiNoProj=hiproc.HiNoProj:main',
            'HiPrecisionInit=hiproc.HiPrecisionInit:main',
            'HiSlither=hiproc.HiSlither:main',
            'HiStitch=hiproc.HiStitch:main',
            'HiccdStitch=hiproc.HiccdStitch:main',
            'JitPlot=hiproc.JitPlot:main',
            'SlitherStats=hiproc.SlitherStats:main',
            'bitflips=hiproc.bitflips:main',
            'fft_clean=hiproc.fft_clean:main',
            'hiproc=hiproc.hiproc:main',
            'lisfix=hiproc.lisfix:main',
            'mdr2cub=hiproc.mdr2cub:main',
            'resolve_jitter=hiproc.resolve_jitter:main',
            'rjplot=hiproc.rjplot:main'
        ],
    },
    install_requires=requirements,
    include_package_data=True,
    package_data={
        "hiproc": ["data/*"],
    },
    keywords='HiRISE',
    packages=find_packages(include=['hiproc', 'hiproc.*']),
    test_suite='tests',
    zip_safe=False,
)



#!/usr/bin/env python
"""Creates special color products ("pretty pictures").

This creates a synthetic RGB product and an enhanced IRB product that
are cosmetically improved over the raw/normalized color products.
HiBeautify takes the .hicolornorm.pvl file (created by HiColorNorm)
as its input. It uses the HiColorNorm cubes (extension _COLOR4.cub
and _COLOR5.cub).

Data Flow
---------
Input Products:

- ``COLOR4.HiColorNorm`` and ``COLOR5.HiColorNorm`` files which are the
    result of HiColorNorm.

Output Products:

- IRB and RGB cubes

"""

# Copyright 2007-2020, Arizona Board of Regents on behalf of the Lunar and
# Planetary Laboratory at the University of Arizona.
#   - Orignal Perl program.
#
# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#   - Elements of this Python program are are based on the original Perl
#     but the logic here is rewritten from scratch to emulate functionality.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This program is based on HiColor version 5.4.3 2020/04/28
# and on the Perl HiBeautify program ($Revision: 1.92 $
#                                     $Date: 2020/04/28 16:56:16 $)
# by Guy McArthur as an employee of the University of Arizona.

import argparse
import logging
import pkg_resources
from datetime import datetime
from pathlib import Path

import pvl

import kalasiris as isis
import hiproc.util as util
import hiproc.HiColorNorm as hcn

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        parents=[util.parent_parser()],
        conflict_handler="resolve",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "-o_irb",
        "--output_irb",
        required=False,
        default="_IRB.cub",
        help="The filename to be used for the output IRB cube.  If it "
             "begins with an underscore ('_') it will be assumed to be a "
             "suffix that will be appended to a name derived from the "
             "observation. Default: %(default)s"
    )
    parser.add_argument(
        "-o_rgb", "--output_rgb",
        required=False,
        default="_RGB.cub",
        help="The filename to be used for the output RGB cube.  If it "
             "begins with an underscore ('_') it will be assumed to be a "
             "suffix that will be appended to a name derived from the "
             "observation. Default: %(default)s"
    )
    parser.add_argument(
        "-c",
        "--conf",
        required=False,
        type=argparse.FileType('r'),
        default=pkg_resources.resource_stream(
            __name__,
            'data/HiBeautify.conf',
        ),
        help="Path to the HiBeautify config file.  Defaults to "
             "HiBeautify.conf distributed with the library."
    )
    # parser.add_argument('-f', '--frost', action='store_true',
    #                     help='Use the frost/ice color stretch, and disable '
    #                     'auto-detection of frost/ice.')
    # parser.add_argument('--nofrost', action='store_true',
    #        help='Do not use the frost/ice color stretch, and disable '
    #                     'auto-detection of frost/ice.')
    parser.add_argument(
        "cubes",
        type=Path,
        nargs=2,
        metavar="HiColorNorm-cube",
        help="The COLOR4.HiColorNorm.cub and COLOR5.HiColorNorm.cub files."
    )
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    with util.main_exceptions(args.verbose):
        HiBeautify(
            args.cubes,
            pvl.load(args.conf),
            args.output_irb,
            args.output_rgb,
            keep=args.keep
        )
    return


def HiBeautify(
    cube_paths: list,
    conf: dict,
    out_irb="_IRB.cub",
    out_rgb="_RGB.cub",
    keep=False
):
    logger.info(f"HiBeautify start: {', '.join(map(str, cube_paths))}")

    # GetConfigurationParameters()
    cubes = list(map(hcn.ColorCube, cube_paths))
    cubes.sort()

    irb_out_p = hcn.set_outpath(out_irb, cubes)
    rgb_out_p = hcn.set_outpath(out_rgb, cubes)

    temp_token = datetime.now().strftime("HiBeautify-%y%m%d%H%M%S")
    to_del = isis.PathSet()

    # Create an IRB mosaic from the HiColorNorm halves.

    # If a half is missing, we create a mosaic with the proper width and place
    # the half in it at the proper location.
    # Logic ofr image_midpoint and total_width come from original HiColorInit
    # irbmerged_p = to_del.add(out_p.with_suffix(f'.{temp_token}_IRB.cub'))
    image_midpoint = int((2000 / cubes[0].red_bin) + 1)
    outsample = image_midpoint
    if cubes[0].ccdnumber == "4":
        outsample = 1
    total_width = int(2 * cubes[0].samps - (48 / cubes[0].red_bin))

    isis.handmos(
        cubes[0].path,
        mosaic=irb_out_p,
        outline=1,
        outsample=outsample,
        outband=1,
        create="Y",
        nlines=cubes[0].lines,
        nsamp=total_width,
        nbands=3,
    )

    if len(cubes) == 1:
        logger.info("Warning, missing one half!")
    else:
        logger.info("Using both halves")
        isis.handmos(
            cubes[1].path,
            mosaic=irb_out_p,
            outline=1,
            outsample=image_midpoint,
            outband=1,
        )

    # Nothing is actually done to the pixels here regarding the FrostStats, so
    # I'm just going to skip them here.
    #
    # # Determine if Frost/ICE may be present using FrostStats module.
    # frost = None
    # if args.frost:
    #     frost = True
    #     logging.info('Frost override: disabling auto-detection and using '
    #                  'the frost/ice color stretch')
    # if args.nofrost:
    #     frost = False
    #     logging.info('Frost override: disable auto-detection and not using '
    #                  'the frost/ice color stretch')
    # if frost is None:
    #     pass
    #     # get frost stats

    # Subtract the unaltered RED band from the high pass filtered BG for
    # synthetic blue.
    logger.info("Creating synthetic B, subtracting RED from BG")
    rgbsynthb_p = to_del.add(irb_out_p.with_suffix(f".{temp_token}_B.cub"))
    isis.algebra(
        f"{irb_out_p}+3",
        from2=f"{irb_out_p}+2",
        op="subtract",
        to=rgbsynthb_p,
        A=conf["Beautify"]["Synthetic_A_Coefficient"],
        B=conf["Beautify"]["Synthetic_B_Coefficient"],
    )

    # Adjust the BandBin group
    isis.editlab(
        rgbsynthb_p, grpname="BandBin", keyword="Name", value="Synthetic Blue"
    )
    isis.editlab(rgbsynthb_p, grpname="BandBin", keyword="Center", value="0")
    isis.editlab(rgbsynthb_p, grpname="BandBin", keyword="Width", value="0")

    # HiBeautify gathers and writes a bunch of statistics to PVL that is
    # important to the HiRISE GDS, but not relevant to just producing pixels
    # so I'm omitting it.
    #
    # # Determine the min and max DN values of each band (RED, IR, BG, B) we're
    # # working with.
    # (upper, lower) = conf['Beautify']['Stretch_Exclude_Lines']
    # if upper == 0 and lower == 0:
    #     synthbcrp_p = rgbsynthb_p
    #     irbmrgcrp_p = irbmerged_p
    # else:
    #     synthbcrp_p = to_del.add(
    #           out_p.with_suffix(f'.{temp_token}_Bx.cub'))
    #     irbmrgcrp_p = to_del.add(
    #           out_p.with_suffix(f'.{temp_token}_IRBx.cub'))
    #
    #     for (f, t) in ((rgbsynthb_p, synthbcrp_p),
    #                    (irbmerged_p, irbmrgcrp_p)):
    #         logging.info(isis.crop(f, to=t, propspice=False,
    #                                line=(1 + upper),
    #                                nlines=(
    #                                   cubes[0].lines - lower + upper)).args)
    #
    # stats = dict()
    # stats['B'] = Get_MinMax(synthbcrp_p,
    #                         conf['Beautify']['Stretch_Reduction_Factor'],
    #                         temp_token, keep=keep)
    #
    # for band in cubes[0].band.keys():
    #     stats[band] = Get_MinMax('{}+{}'.format(str(irbmrgcrp_p),
    #                                             cubes[0].band[band]),
    #                              conf['Beautify']['Stretch_Reduction_Factor'],
    #                              temp_token, keep=keep)

    # Create an RGB cube using the RED from the IRB mosaic,
    # the BG from the IRB mosaic and the synthetic B that we just made.
    isis.cubeit_k(
        [f"{irb_out_p}+2", f"{irb_out_p}+3", rgbsynthb_p], to=rgb_out_p
    )

    if not keep:
        to_del.unlink()

    return


# Turns out that HiBeautify doesn't really need this functionality.  This was
# just development code and hasn't been tested.
#
# def Get_MinMax(cube, scale=1,
#                temp_token=datetime.now().strftime('%y%m%d%H%M%S'),
#                keep=False) -> tuple:
#     '''Reduces an Isis cub by a specified scaling factor.  Runs hist to
#        return the pixel values from the histogram at the ordinal
#        positions passed (or min & max if unspecified).
#
#        Returns (min, max, avg).
#     '''
#     # Taken from HiArch, not HiBeautify, but seems to be called here first.
#
#     in_p = Path(cube)
#     to_del = isis.PathSet()
#
#     if scale > 1:
#         reduce_p = to_del.add(in_p.with_suffix(f'{temp_token}.reduce.cub'))
#         logging.info(isis.reduce(in_p, to=reduce_p, sscale=scale,
#                                  lscale=scale, validper=1,
#                                  vper_replace='nearest', mode='SCALE').args)
#
#         in_p = reduce_p
#
#     elif scale < 1:
#         raise ValueError(
#             f'The value for scale must be >= 1, but was {scale}.')
#
#     stats = isis.stats_k(in_p)
#
#     if not keep:
#         to_del.unlink()
#
#     if int(stats['NullPixels']) >= int(stats['TotalPixels']):
#         return (-1, -1, 0)
#
#     return(int(stats['Minimum']), int(stats['Maximum']),
#            int(stats['Average']))



#!/usr/bin/env python
"""Perform fast-fourier-transform cleaning on HiRISE Channels."""

# Copyright 2021, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This program is based on procedures and code written by
# Audrie Fennema and Sarah Mattson circa April 2013.

import argparse
import logging
import os
import shutil
import sys
from datetime import datetime
from pathlib import Path

import pvl
import kalasiris as isis

import hiproc.util as util

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[util.parent_parser()],
    )
    parser.add_argument(
        "-o", "--output",
        required=False,
        default=".fftclean.cub",
        help="The name of the output .cub file to write.  Optionally, if "
             "it starts with a '.' it is considered a suffix "
             "and the input file will have its suffix removed, and this will "
             "be added to input file name. Default: %(default)s",
    )
    parser.add_argument("file", help="ISIS Cube file to clean.")
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    out_p = util.path_w_suffix(args.output, args.file)

    with util.main_exceptions(args.verbose):
        clean(
            args.file,
            out_p,
            keep=args.keep,
        )
    sys.exit(0)


def clean(
    in_path: os.PathLike,
    out_path: os.PathLike,
    keep=False,
):
    """The file at *out_path* will be the result of running fft
    cleaning of the file at *in-path*.

    If *keep* is True, then all intermediate files will be preserved,
    otherwise, this function will clean up any intermediary files
    it creates.
    """
    in_p = Path(in_path)
    out_p = Path(out_path)

    temp_token = datetime.now().strftime("fftclean-%y%m%d%H%M%S")
    to_delete = isis.PathSet()

    # Check input channel cube for gaps. If they exist, interpolate and run
    # fft_clean on the interpolated gaps cube. Replace gaps in final cleaned
    # image.
    gaplog = to_delete.add(in_p.with_suffix(f".{temp_token}.gaplog.pvl"))
    isis.findgaps(in_p, log=gaplog)
    gappvl = pvl.load(gaplog)
    next_cube = in_p
    if "Gap" in gappvl:
        logger.info(f"{in_p.name} has gaps.")
        next_cube = to_delete.add(
            in_p.with_suffix(f".{temp_token}.fillgap.cub")
        )
        isis.fillgap(in_p, to=next_cube, interp="linear", direction="sample")
    else:
        logger.info(f"{in_p.name} has no gaps.")

    # Take the FFT of the input channel cube
    magcube = to_delete.add(in_p.with_suffix(f".{temp_token}.mag.cub"))
    phasecube = to_delete.add(in_p.with_suffix(f".{temp_token}.pha.cub"))
    isis.fft(next_cube, magnitude=magcube, phase=phasecube)

    samps = int(isis.getkey_k(magcube, "Dimensions", "Samples"))
    lines = int(isis.getkey_k(magcube, "Dimensions", "Lines"))

    # Make the filter for the magnitude image
    # Make circle mask based on butterworth filter idea (changed /45 to /25)
    circ = to_delete.add(in_p.with_suffix(f".{temp_token}.circ.cub"))
    isis.fx(
        to=circ,
        equation="1 - (1/{--pi*[1+((line - 51)/45)^2]} + "
                 "1/{--pi*[1 + ((sample - 51)/45)^2]})",
        mode="outputonly",
        lines=101,
        samples=101
    )
    circ_str = to_delete.add(
        in_p.with_suffix(f".{temp_token}.circ.stretch.cub")
    )
    isis.stretch(
        circ,
        to=circ_str,
        usepercentages=True,
        pairs="0.0:1 0.01:1 75:.99 80:0.9 100:0"
    )
    circ_enl = to_delete.add(
        in_p.with_suffix(f".{temp_token}.circ.enl.cub")
    )
    isis.enlarge(
        circ_str,
        to=circ_enl,
        interp="cubic",
        mode="total",
        ons=samps,
        onl=lines
    )

    # Make full size cross mask
    crossfull = to_delete.add(
        in_p.with_suffix(f".{temp_token}.cross_full.cub")
    )
    isis.fx(
        to=crossfull,
        equation=f"[1 - e^(--pi*[(line - (( {lines} / 2 ) +1))^6])] * "
                 f"[1 - e^(--pi*[(sample - (( {samps} / 2) +1))^6])]",
        mode="outputonly",
        lines=lines,
        samples=samps
    )

    maskcube = to_delete.add(
        in_p.with_suffix(f".{temp_token}.mask.cub")
    )
    isis.fx(
        to=maskcube,
        f1=crossfull,
        f2=circ_enl,
        equation="f1 * f2"
    )

    # Filter the mag image before detecting spikes
    # (filter out low frequency information)
    magfilt = in_p.with_suffix(f".{temp_token}.magfilt.cub")
    isis.fx(f1=magcube, f2=maskcube, to=magfilt, equation="f2 * abs(f1)")

    # Normalize the filtered magnitude image
    magnorm = float(pvl.loads(isis.stats(magfilt).stdout)["Results"]["Median"])

    magfilt_cn = in_p.with_suffix(f".{temp_token}.magfilt.cubenorm.cub")
    isis.fx(f1=magfilt, to=magfilt_cn, equation=f"f1/{magnorm}")

    # Run a Sobel (gradient) filter on the filtered magnitude image to
    # emphasize spikes.
    magfilt_sobel = in_p.with_suffix(f".{temp_token}.magfilt.sobel.cub")
    isis.gradient(magfilt_cn, to=magfilt_sobel, gradtype="sobel")

    # Threshold the normalized, filtered mag image to find spikes.
    # N times average empirically derived value. Typically n=5
    threshold = 5
    threshcube = in_p.with_suffix(f".{temp_token}.threshold.cub")
    isis.fx(
        f1=magfilt_sobel,
        to=threshcube,
        equation=f"f1 > ( {threshold} * cubeavg(f1) )"
    )

    # Use a Gaussian filter to broaden the spikes.
    gaus_cube = in_p.with_suffix(f".{temp_token}.threshold.gaus.cub")
    isis.gauss(threshcube, to=gaus_cube, size=5, stddev=6.0)

    # Threshold the Gaussian filtered cube and re-apply the cross mask to it.
    # This is to re-exclude the cross which the Gaussian filter might have
    # spread into.
    thresh2 = in_p.with_suffix(f".{temp_token}.threshold2.cub")
    isis.fx(f1=gaus_cube, to=thresh2, equation=" f1 > 0 ")

    wcross = in_p.with_suffix(f".{temp_token}.threshold2.wcross.cub")
    isis.fx(f1=thresh2, f2=crossfull, to=wcross, equation=" f1 * f2 ")

    # Multiply the binary threshold image by the magnitude maximum, plus 1.
    magmax = float(pvl.loads(isis.stats(magcube).stdout)["Results"]["Maximum"])
    magfilt_divby = in_p.with_suffix(f".{temp_token}.magfilt.divideby.cub")
    isis.fx(f1=wcross, to=magfilt_divby, equation=f"1 + (f1 * {magmax})")

    # Divide the unfiltered magnitude image by the thresholded image to
    # minimize the noise spikes.
    magclean = in_p.with_suffix(f".{temp_token}.magclean.cub")
    isis.fx(f1=magcube, f2=magfilt_divby, equation="f1 / f2", to=magclean)

    # Perform the inverse transform with the cleaned magnitude image and the
    # original phase image.
    cleaned = in_p.with_suffix(f".{temp_token}.clean.cub")
    isis.ifft(magnitude=magclean, phase=phasecube, to=cleaned)

    if "Gap" in gappvl:
        gapclean = in_p.with_suffix(f".{temp_token}.cleangapped.cub")
        isis.fx(f1=cleaned, f2=in_p, to=gapclean, equation=" f1 + (0 * f2 )")
        cleaned = gapclean

    # Subtract the cleaned image from the cubenormed image to make a
    # difference cube. Gather stats.
    diff = in_p.with_suffix(f".{temp_token}.diff.cub")
    isis.algebra(in_p, from2=cleaned, to=diff, operator="subtract")
    # Command("stats from=$base.cubenorm.cub");
    # Command("stats from=$base.clean.cub");
    # Command("stats from=$base.diff.cub > $base.diff.stats.txt");

    # Command("/bin/rm -f $base.magfilt.divideby.cub $base.magfilt.cubenorm.cub");

    # Make a crop of the difference cubes for validation.
    crop_line_start = int((lines / 2) - (samps / 2))
    diff_crop = in_p.with_suffix(f".{temp_token}.diff.crop.cub")
    isis.crop(
        diff,
        to=diff_crop,
        sample=1,
        nsamples=samps,
        line=crop_line_start,
        nline=samps
    )

    shutil.copy(cleaned, out_p)

    if not keep:
        to_delete.unlink()

    return


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Deal with stuck bits in HiRISE pixels via unflipping.

This program will attempt to unflip the bit-flipped pixels.
Only the image area is affected. This is also an older approach
and is less supported than bitflips.py.

Please see bitflips.py for a more detailed explanation of
the HiRISE bit-flip problem.

This file contains functions to deal with 'unflipping' the bits.
"""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Parts of this were inspired by clean_bit_flips.pro by Alan Delamere,
# Oct 2019, but the implementation here was written from scratch.

import argparse
import itertools
import logging
import math
import shutil
import subprocess
import sys
from pathlib import Path

import kalasiris as isis

import hiproc.util as util

logger = logging.getLogger(__name__)


def main():
    try:
        parser = argparse.ArgumentParser(
            description=__doc__,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            parents=[util.parent_parser()],
        )
        parser.add_argument(
            "-o", "--output", required=False, default=".bitunflip.cub"
        )
        parser.add_argument("cube", help="ISIS Cube file.")

        args = parser.parse_args()

        util.set_logger(args.verbose, args.logfile, args.log)

        out_p = util.path_w_suffix(args.output, args.cube)

        unflip(Path(args.cube), out_p, keep=args.keep)

    except subprocess.CalledProcessError as err:
        print("Had an ISIS error:", file=sys.stderr)
        print(" ".join(err.cmd), file=sys.stderr)
        print(err.stdout, file=sys.stderr)
        print(err.stderr, file=sys.stderr)


def get_range(start, stop, step=None) -> range:
    """Returns a range object given floating point start and
    stop values (the optional step must be an int).  If
    step is none, it will be set to 1 or -1 depending on the
    relative values of start and stop.
    """
    if start < stop:
        start = math.floor(start)
        stop = math.ceil(stop)
        if step is None:
            step = 1
    else:
        start = math.ceil(start)
        stop = math.floor(stop)
        if step is None:
            step = -1

    return range(start, stop, step)


def find_gap(hist: list, start, stop, step=None, findfirst=False) -> int:
    """Returns a DN between *start* and *stop* which is determined to be a gap.

    Given the range specified by *start*, *stop* and *step*, the
    DNs are extracted from the list of namedtuples (which must
    conform to the HistRow namedtuple from kalasiris.Histogram).
    The 'gaps' between continuous ranges of DN are found (DNs where
    the histogram has no pixels).  The largest gap is found, and
    then the DN value closest to start from that largest gap is
    returned.

    If step is not specified or None, it will be set to 1 or -1
    depending on the relative values of start and stop.

    If *findfirst* is `True`, then rather than finding the 'biggest'
    gap, it will return the DN from the gap that is closest to
    *start*.
    """
    dn_window = get_range(start, stop, step)

    hist_set = set(filter(lambda d: d in dn_window, (int(x.DN) for x in hist)))

    missing = sorted(
        set(dn_window).difference(hist_set), reverse=(dn_window.step < 0)
    )
    # logging.info(f'{bs} missing: ' + str(missing))

    if len(missing) > 0:
        sequences = list()
        for k, g in itertools.groupby(
            missing,
            (lambda x, c=itertools.count(): next(c) - (dn_window.step * x)),
        ):
            sequences.append(list(g))

        if findfirst:
            return sequences[0][0]
        else:
            # find biggest gap
            return max(sequences, key=len)[0]
    else:
        # There was no gap in DN.
        raise ValueError(
            "There was no gap in the DN window from " f"{start} to {stop}."
        )


def find_min_dn(hist: list, start, stop, step=None) -> int:
    """Returns the DN from *hist* with the lowest pixel count.

    Where *hist* is a list of namedtuples (which must conform to
    the HistRow namedtuple from kalasiris.Histogram).  The *start*,
    *stop*, and *step* parameters are DN values in that *hist* list.

    Given how the min() mechanics work, this should return the first
    DN value of the entry with the lowest pixel count, in range order
    (if there is more than one).
    """
    r = get_range(start, stop, step)
    hist_list = sorted(
        filter(lambda x: int(x.DN) in r, hist),
        key=lambda x: int(x.DN),
        reverse=(r.step < 0),
    )

    h = min(hist_list, key=lambda x: int(x.Pixels))
    return int(h.DN)


def subtract_over_thresh(
    in_path: Path, out_path: Path, thresh: int, delta: int, keep=False
):
    """This is a convenience function that runs ISIS programs to add or
    subtract a value to DN values for pixels that are above or below
    a threshold.

    For all pixels in the *in_path* ISIS cube, if *delta* is positive,
    then pixels with a value greater than *thresh* will have *delta*
    subtracted from them.  If *delta* is negative, then all pixels
    less than *thresh* will have *delta* added to them.
    """

    # Originally, I wanted to just do this simply with fx:
    # eqn = "\(f1 + ((f1{glt}={thresh}) * {(-1 * delta)}))"
    # However, fx writes out floating point pixel values, and we really
    # need to keep DNs as ints as long as possible.  Sigh.

    shutil.copyfile(in_path, out_path)

    mask_p = in_path.with_suffix(".threshmask.cub")
    mask_args = {"from": in_path, "to": mask_p}
    if delta > 0:
        mask_args["min"] = thresh
    else:
        mask_args["max"] = thresh
    isis.mask(**mask_args)

    delta_p = in_path.with_suffix(".delta.cub")
    isis.algebra(
        mask_p, from2=in_path, to=delta_p, op="add", a=0, c=(-1 * delta)
    )

    isis.handmos(delta_p, mosaic=out_path)

    if not keep:
        mask_p.unlink()
        delta_p.unlink()

    return


def mask_gap(in_path: Path, out_path: Path):
    """Attempt to mask out pixels beyond the central DNs of the median
    based on gaps in the histogram.

    This approach worked well based on 'ideal' reverse-clocked data
    or 'dark' images, but in 'real' HiRISE images of Mars, the reality
    is that there are 'gaps' everywhere along the DN range, and this
    approach ends up being too 'dumb'.
    """

    hist = isis.Histogram(in_path)

    median = math.trunc(float(hist["Median"]))
    # std = math.trunc(float(hist['Std Deviation']))

    high = find_gap(
        hist, median, math.trunc(float(hist["Maximum"])), findfirst=True
    )

    low = find_gap(
        hist, median, math.trunc(float(hist["Minimum"])), findfirst=True
    )

    highdist = high - median
    lowdist = median - low

    maskmax = find_gap(hist, median, (median + (2 * highdist)))
    maskmin = find_gap(hist, median, (median - (2 * lowdist)))

    isis.mask(in_path, to=out_path, minimum=maskmin, maximum=maskmax)

    return


def get_unflip_thresh(hist: list, far, near, lowlimit) -> int:
    """Provides a robust threshhold for the unflipping process."""
    # try:
    thresh = find_gap(hist, far, near)
    # except ValueError:
    #     logging.info('No zero threshold found.')
    #     # d is not defined here, and I think this was the victim of
    #     # some cut'n'paste work, but would need to go analyze the
    #     # repo history to track it down.  For the moment, this will
    #     # just error!
    #     if d >= lowlimit:
    #         thresh = find_min_dn(hist, far, near)
    #         logging.info(f'Found a minimum threshold: {thresh}')
    #         if thresh == far or thresh == near:
    #             raise ValueError(f'Minimum, {thresh}, at edge of range.')
    #     else:
    #         raise ValueError(f'Below {lowlimit}, skipping.')
    return thresh


def unflip(in_p: Path, out_p: Path, keep=False):
    """Attempt to indentify DNs whose bits have been flipped in the
    ISIS cube indicated by *in_p*, and unflip them.

    Only operates on image-area.
    """
    to_del = isis.PathSet()

    # This full suite of deltas works well for the reverse-clock area
    # and even 'dark' images, but not 'real' images.
    # deltas = (8192, 4096, 2048, 1024, 512, 256, 128, 64)

    # Restricting the number of deltas might help, but this seems
    # arbitrary.
    deltas = (8192, 4096, 2048, 1024)

    count = 0
    suffix = ".bf{}-{}{}.cub"
    this_p = to_del.add(in_p.with_suffix(suffix.format(count, 0, 0)))
    this_p.symlink_to(in_p)

    median = math.trunc(float(isis.stats_k(in_p)["Median"]))

    for (sign, pm, extrema) in ((+1, "m", "Maximum"), (-1, "p", "Minimum")):
        # logging.info(pm)
        for delt in deltas:
            d = sign * delt
            far = median + d
            near = median + (d / 2)

            hist = isis.Histogram(this_p)

            logger.info(
                f"bitflip position {pm}{delt}, near: {near} "
                f"far: {far}, extrema: {hist[extrema]}"
            )
            if (sign > 0 and far < float(hist[extrema])) or (
                sign < 0 and far > float(hist[extrema])
            ):
                count += 1
                s = suffix.format(count, pm, delt)
                next_p = to_del.add(this_p.with_suffix("").with_suffix(s))
                try:
                    thresh = get_unflip_thresh(hist, far, near, d)
                except ValueError as err:
                    logger.info(err)
                    count -= 1
                    break
                subtract_over_thresh(this_p, next_p, thresh, d, keep=keep)
                this_p = next_p
            else:
                logger.info(
                    "The far value was beyond the extrema. " "Didn't bother."
                )

    shutil.move(this_p, out_p)

    if not keep:
        to_del.remove(this_p)
        to_del.unlink()

    return


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Stitch together two HiRISE channel files to create a single CCD image file.

These are the functionalities that the HiStitch pipeline performs that are
reproduced here:

- Join (or stitch) together two channel products to form a whole CCD product.
    The channel 1 product is on the left, channel 0 on the right.
- When joining the two channel products, the ISIS histitch program first
    obtains the image average of the two channel products at the
    join area. The program then adjusts the average of channel 1
    to match the average of channel 0 through a multiplicative
    correction. This step produces a radiometrically seamless CCD
    product.
- For Bin 2 and 4 images a fix is applied to images that have experienced
    furrowing. If furrowing has been detected, a highpass/lowpass
    filter is applied to the furrow region for those pixels that
    have not been set to the null pixel value in the HiCal
    step.

HiStitch is the third step in the HiRISE processing chain, after HiCal and
results in data that is read for the next step: HiccdStitch.  If you are
running HiCal individually without all EDRs from the Observation in the
same directory, please be aware of the --bin2 and --bin4 options.


Data Flow
---------
Input Products:

- Two ``.cub`` files from the same CCD that have been processed through HiCal.
- Two ``.json`` files that contain summary information about the .cub files.

Output Products:

- A stitched ``.cub`` file.
- A ``.json`` file with summary information about the stitched image.

"""

# Copyright 2004-2020, Arizona Board of Regents on behalf of the Lunar and
# Planetary Laboratory at the University of Arizona.
#   - Orignal Perl program.
#
# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#   - Elements of this Python program are are based on the original Perl
#     but the logic here is rewritten from scratch to emulate functionality.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This program is based on HiStitch version 2.21.1 (2021/01/01),
# and on the Perl HiStitch program ($Revision: 1.26 $
#                                   $Date: 2020/04/28 17:52:53 $)
# and on the Perl HiFurrow_Fix program ($Revision: 1.12 $
#                                       $Date: 2020/10/30 00:48:28 $)
# by Eric Eliason as an employee of the University of Arizona.

import argparse
import collections
import json
import logging
import os
import pkg_resources
import shutil
from datetime import datetime
from pathlib import Path

import pvl

import kalasiris as isis
import hiproc.hirise as hirise
import hiproc.util as util

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        parents=[util.parent_parser()],
        conflict_handler="resolve",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "--db",
        required=False,
        default=".HiCat.json",
        help="The .json file to use.  Optionally, if it "
        "starts with a '.' it is considered an extension and "
        "will be swapped with the input file's extension to "
        "find the .json file to use. Default: %(default)s",
    )
    parser.add_argument(
        "--db2",
        required=False,
        default=".HiCat.json",
        help="The second .json file to use.  Optionally, if "
        "it starts with a '.' it is considered a suffix "
        "and will be swapped with the second input file's "
        "suffix to find the .json file to use. Default: %(default)s",
    )
    parser.add_argument(
        "--dbout",
        required=False,
        default=".HiCat.json",
        help="The name of the output .json file to write.  Optionally, if "
             "it starts with a '.' it is considered a suffix "
             "and will be added to the CCD ID of the observation of the "
             "input files. Default: %(default)s",
    )
    parser.add_argument(
        "-o", "--output",
        required=False,
        default=".HiStitch.cub",
        help="The name of the output .cub file to write.  Optionally, if "
             "it starts with a '.' it is considered a suffix"
             "and will be added to the CCD ID of the observation of the "
             "input files. Default: %(default)s",
    )
    parser.add_argument(
        "-c",
        "--conf",
        required=False,
        type=argparse.FileType('r'),
        default=pkg_resources.resource_stream(
            __name__,
            'data/HiStitch.conf'
        ),
        help="Path to the HiStitch config file.  Defaults to "
             "HiStitch.conf distributed with the library."
    )
    parser.add_argument("cube0", metavar="cube0.cub-file")
    parser.add_argument("cube1", metavar="cube1.cub-file", nargs="?")
    return parser


def main():
    # The original Perl needed the specific output of cubenorm from a
    # particular step in the HiCal pipeline before here.  However, rather
    # than keep track of those files, open them, and read them again, we
    # can (and did) perform the relevant check in HiCal.py, and then save
    # that in the db that we can check now.

    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    with util.main_exceptions(args.verbose):
        (db, outcub_path) = HiStitch(
            args.cube0,
            args.cube1,
            get_db(args.db, args.cube0),
            get_db(args.db2, args.cube1),
            args.output,
            pvl.load(args.conf),
            keep=args.keep,
        )

    db_path = set_outpath(
        args.dbout, hirise.get_CCDID_fromfile(outcub_path), outcub_path.parent
    )

    with open(db_path, "w") as f:
        json.dump(db, f, indent=0, sort_keys=True)

    logger.info(f"Wrote {db_path}")
    return


def get_db(db: str, cube_path: os.PathLike) -> dict:
    if cube_path is None:
        return None

    db_path = util.pid_path_w_suffix(db, cube_path)

    with open(db_path) as f:
        db = json.load(f)

    return db


def get_chids(cubes: list) -> tuple:

    cid_a = hirise.get_ChannelID_fromfile(cubes[0])

    if len(cubes) == 2:
        cid_b = hirise.get_ChannelID_fromfile(cubes[1])
        if hirise.CCDID(cid_a) == hirise.CCDID(cid_b):
            return (cid_a, cid_b)
        else:
            raise ValueError(
                "These do not appear to be channels "
                f"from the same CCD:\n{cubes[0]}: "
                f"{cid_a} and\n{cubes[1]}: {cid_b}"
            )
    else:
        return (cid_a,)


def set_outpath(
    output: os.PathLike, cid: hirise.CCDID, directory: Path
) -> Path:
    if str(output).startswith("."):
        return directory / Path(str(cid) + output)
    else:
        return Path(output)


def sort_input_cubes(a_cub: os.PathLike, b_cub: os.PathLike) -> tuple:
    """Figures out which one is Channel 0 and which is Channel 1."""

    channel_of = dict()
    channel_of[a_cub] = isis.getkey_k(a_cub, "Instrument", "ChannelNumber")
    channel_of[b_cub] = isis.getkey_k(b_cub, "Instrument", "ChannelNumber")

    if channel_of[a_cub] == channel_of[b_cub]:
        raise RuntimeError(
            f"{a_cub} and {b_cub} have the same channel: "
            "{channel_of[a_cub]}."
        )

    for (cub, chan) in channel_of.items():
        if int(chan) != 0 and int(chan) != 1:
            raise RuntimeError(
                f"{cub} has a channel other than 0 or 1: {chan}"
            )

    sorted_cubchan = sorted(channel_of.items(), key=lambda x: x[1])
    return (sorted_cubchan[0][0], sorted_cubchan[1][0])


def sort_databases(dbs: list, chids: tuple) -> tuple:
    """Ensures that the databases match the cubes."""

    if len(dbs) != len(chids):
        raise IndexError(
            f"The number of databases ({len(dbs)})  does not "
            f"match the number of Product IDs ({chids})"
        )

    chid_db_map = dict()
    for p in chids:
        for d in dbs:
            if str(p) == d["PRODUCT_ID"]:
                chid_db_map[p] = d
                break
    if len(chid_db_map) != len(chids):
        raise LookupError(
            f"The Product IDs in the databases do not match {chids}"
        )
    return tuple(map(chid_db_map.get, chids))


def HiStitch(
    cube0: os.PathLike,
    cube1: os.PathLike,
    db0: dict,
    db1: dict,
    out_cube: os.PathLike,
    conf: dict,
    keep=False,
) -> tuple:
    logger.info(f"HiStitch start: {cube0} {cube1}")

    # GetConfigurationParameters()
    conf_check(conf)

    # GetProductFiles()
    if not cube1:
        cubes = (Path(cube0),)
        db_list = [db0, ]
    else:
        cubes = sort_input_cubes(Path(cube0), Path(cube1))
        db_list = [db0, db1]

    chids = get_chids(cubes)

    out_path = set_outpath(out_cube, hirise.CCDID(chids[0]), cubes[0].parent)

    # PrepareDBStatements()
    #   select from HiCat.EDR_Products
    db_paths = list()
    db_paths.append(db0)
    if len(cubes) == 2:
        db_paths.append(db1)

    dbs = sort_databases(db_list, chids)
    ccd_number = int(chids[0].ccdnumber)

    # Allows for indexing in lists ordered by bin value.
    b = 1, 2, 4, 8, 16

    # This string will get placed in the filename for all of our
    # temporary files. It will (hopefully) prevent collisions with
    # existing files and also allow for easy clean-up if keep=True
    temp_token = datetime.now().strftime("HiStitch-%y%m%d%H%M%S")

    # ProcessingStep() - mostly sets up stuff
    max_mean = max(map(lambda x: float(x["IMAGE_MEAN"]), dbs))
    flags = set_flags(
        conf["HiStitch"],
        dbs,
        ccd_number,
        b.index(int(dbs[0]["BINNING"])),
        max_mean
    )

    # HiStitchStep() - runs HiStitch, and originally inserted to db, now we
    # just return at the bottom of this function.
    logger.info(
        HiStitchStep(
            cubes,
            out_path,
            dbs[0]["BINNING"],
            ccd_number,
            conf["HiStitch"],
            flags.balance,
            flags.equalize,
        ).args
    )

    if len(cubes) == 2 and flags.balance:
        # run getkey from ?HiStitch_output?
        truthchannel = isis.getkey_k(out_path, "HiStitch", "TruthChannel")
        balanceratio = isis.getkey_k(out_path, "HiStitch", "BalanceRatio")
    else:
        truthchannel = None
        balanceratio = None

    if flags.furrow:
        furrow_file = out_path.with_suffix(f".{temp_token}.temp.cub")
        HiFurrow_Fix(out_path, furrow_file, max_mean, keep=keep)
        furrow_file.rename(out_path)

    # insert ObsID, pid0.ccdname+pid0.ccdnumber, truthchannel, balanceratio
    # into HiCat.CCD_Processing_Statistics
    db = dict()
    db["OBSERVATION_ID"] = str(chids[0].get_obsid())
    db["CCD"] = chids[0].get_ccd()
    db["CONTROL_CHANNEL"] = truthchannel
    db["CHANNEL_MATCHING_CORRECTION"] = balanceratio

    logger.info(f"HiStitch done: {out_path}")
    return db, out_path


def set_flags(
    conf, dbs, ccdnum: int, bindex: int, max_mean: float
) -> collections.namedtuple:
    """Set various processing flags based on various configuration
    parameters."""
    HiStitchFlags = collections.namedtuple(
        "HiStitchFlags", ["furrow", "balance", "equalize"]
    )

    max_lis = max(map(lambda x: int(x["LOW_SATURATED_PIXELS"]), dbs))
    max_std = max(map(lambda x: float(x["CAL_MASK_STANDARD_DEVIATION"]), dbs))
    max_dstd = max(
        map(lambda x: float(x["IMAGE_DARK_STANDARD_DEVIATION"]), dbs)
    )
    max_gper = max(map(lambda x: float(x["GAP_PIXELS_PERCENT"]), dbs))

    binning = int(dbs[0]["BINNING"])

    # We don't run FurrowCheck() here like the original Perl does, but do it
    # earlier in HiCal.py and store the result to get now.

    # Determine if the Furrow Normalization is to occur
    furrow = False
    if (
        not any(map(lambda x: bool(x["zapped"]), dbs))
        and max_mean >= float(conf["HiStitch_Furrow_Normalization"])
        and (binning == 2 or binning == 4)
    ):
        furrow = True

    # Determine if the balance or equalize options are to occur
    balance = False
    equalize = False

    if conf["HiStitch_Balance"] or conf["HiStitch_Equalize"]:
        if int(conf["HiStitch_Balance_Processing"][ccdnum]) == 0 or (
            max_lis < int(conf["HiStitch_LIS_Pixels"])
            # The original Perl had these two following conditions written
            # this way, but I think the lists need to be indexed by bindex:
            and max_std < float(conf["HiStitch_Balance_Bin_Mask_STD"][binning])
            and max_dstd
            < float(conf["HiStitch_Balance_Bin_DarkPixel_STD"][binning])
        ):
            logger.warning(
                "Original Perl issue: conf file arrays are being "
                "indexed incorrectly, may affect setting of "
                "equalize and balance flags."
            )
            # To correct, change the indexing of the conf lists above with
            # 'bindex' instead of 'binning'.

            # equalize and balance can't both be true
            if conf["HiStitch_Equalize"] and max_gper < float(
                conf["HiStitch_Gap_Percent"]
            ):
                equalize = True
            elif conf["HiStitch_Balance"]:
                balance = conf["HiStitch_Balance"]

    return HiStitchFlags(furrow, balance, equalize)


def HiStitchStep(in_cubes, out_cub, binning, ccdnum, conf, balance, equalize):
    zpad_bin = "{:0>2}".format(binning)
    histitch_args = {"from1": in_cubes[0], "to": out_cub}
    if len(in_cubes) == 2:
        histitch_args["from2"] = in_cubes[1]
    if not balance and not equalize:
        histitch_args["balance"] = False
    if equalize or balance:
        histitch_args["skip"] = conf[f"HiStitch_Bin{zpad_bin}_Area"][0]
        histitch_args["seamsize"] = conf[f"HiStitch_Bin{zpad_bin}_Area"][1]
        histitch_args["channel"] = conf["HiStitch_Control_Channel"][ccdnum]
        if balance:
            histitch_args["balance"] = "TRUE"
        if equalize:
            histitch_args["balance"] = "EQUALIZE"
            histitch_args["width"] = conf["HiStitch_Equalize_Width"]
            histitch_args["operator"] = conf["HiStitch_Equalize_Correction"]

    return isis.histitch(**histitch_args)


def HiFurrow_Fix(
    in_cube: os.PathLike, out_cube: os.PathLike, max_mean: float, keep=False
):
    """Perform a normalization of the furrow region of bin 2 or 4
    HiRISE images. The input to this script is a HiRISE stitch
    product containing both channels of a CCD.
    """
    in_cub = Path(in_cube)

    binning = int(isis.getkey_k(in_cub, "Instrument", "Summing"))
    lines = int(isis.getkey_k(in_cub, "Dimensions", "Lines"))
    samps = int(isis.getkey_k(in_cub, "Dimensions", "Samples"))

    if binning != 2 and binning != 4:
        raise ValueError(
            "HiFurrow_Fix only supports correction for " "bin 2 or 4 data."
        )
    if binning == 2 and samps != 1024:
        raise ValueError(
            f"HiFurrowFix: improper number of samples: {samps}, "
            "for a stitch product with bin 2 (should be 1024)."
        )

    # This string will get placed in the filename for all of our
    # temporary files. It will (hopefully) prevent collisions with
    # existing files and also allow for easy clean-up if keep=True
    temp_token = datetime.now().strftime("HFF-%y%m%d%H%M%S")
    to_del = isis.PathSet()

    # For bin2 and bin4 imaging, specify width of furrow based on
    # image average DN range
    range_low = {2: (512, 513), 4: (256, 257)}  # 2 pixel furrow width
    range_mid = {2: (511, 514), 4: (255, 258)}  # 4 pixel furrow width
    range_hgh = {2: (511, 514), 4: (255, 258)}  # 4 pixel furrow width
    range_max = {2: (510, 515), 4: (254, 259)}  # 6 pixel furrow width

    # Original code had low/mid/hgh for bin2 and bin4, but they
    # were hard-coded to be identical.
    dn_range_low = 9000
    dn_range_mid = 10000
    dn_range_hgh = 12000

    if max_mean > dn_range_hgh:
        dn_range = range_max[binning]
    elif max_mean > dn_range_mid:
        dn_range = range_hgh[binning]
    elif max_mean > dn_range_low:
        dn_range = range_mid[binning]
    else:
        dn_range = range_low[binning]

    lpf_samp = int((dn_range[1] - dn_range[0] + 1) / 2) * 4 + 1
    lpf_line = int(lpf_samp / 2) * 20 + 1

    # Create a mask file
    # DN=1 for non-furrow area
    # DN=0 for furrow area
    eqn = rf"\(1*(sample<{dn_range[0]})+ 1*(sample>{dn_range[1]}) + 0)"
    fx_cub = to_del.add(in_cub.with_suffix(f".{temp_token}.fx.cub"))
    isis.fx(
        to=fx_cub, mode="OUTPUTONLY", lines=lines, samples=samps, equation=eqn
    )

    # Create a file where the furrow area is set to null
    mask1_cub = to_del.add(in_cub.with_suffix(f".{temp_token}.mask1.cub"))
    isis.mask(
        in_cub,
        mask=fx_cub,
        to=mask1_cub,
        min_=1,
        max_=1,
        preserve="INSIDE",
        spixels="NULL",
    )

    # Lowpass filter to fill in the null pixel area
    lpf_cub = to_del.add(in_cub.with_suffix(f".{temp_token}.lpf.cub"))
    isis.lowpass(
        mask1_cub,
        to=lpf_cub,
        sample=lpf_samp,
        line=lpf_line,
        null=True,
        hrs=False,
        his=False,
        lis=False,
    )

    # Create a file where non-furrow columns are set to null
    mask2_cub = to_del.add(in_cub.with_suffix(f".{temp_token}.mask2.cub"))
    isis.mask(
        in_cub,
        mask=fx_cub,
        to=mask2_cub,
        min_=0,
        max_=0,
        preserve="INSIDE",
        spixels="NULL",
    )

    # Highpass filter the furrow region
    hpf_cub = to_del.add(in_cub.with_suffix(f".{temp_token}.hpf.cub"))
    isis.highpass(mask2_cub, to=hpf_cub, sample=1, line=lpf_line)

    # Add lowpass and highpass together to achieve desired result
    alg_cub = to_del.add(in_cub.with_suffix(f".{temp_token}.alg.cub"))
    isis.algebra(
        from_=lpf_cub, from2=hpf_cub, to=alg_cub, operator="ADD", A=1.0, B=1.0
    )

    # copy the input file to the output file then mosaic the
    # furrow area as needed.
    logger.info(f"Copy {in_cub} to {out_cube}.")
    shutil.copyfile(in_cub, out_cube)
    isis.handmos(
        alg_cub,
        mosaic=out_cube,
        outsample=1,
        outline=1,
        outband=1,
        insample=1,
        inline=1,
        inband=1,
        create="NO",
    )

    if not keep:
        to_del.unlink()

    return


def conf_check(conf: dict) -> None:
    """Various checks on parameters in the configuration."""

    util.conf_check_strings(
        "HiStitch_Clean_Files",
        ("DELETE", "KEEP"),
        conf["HiStitch"]["HiStitch_Clean_Files"],
    )

    util.conf_check_bool(
        "HiStitch_Balance", conf["HiStitch"]["HiStitch_Balance"]
    )

    util.conf_check_bool(
        "HiStitch_Equalize", conf["HiStitch"]["HiStitch_Equalize"]
    )

    util.conf_check_strings(
        "HiStitch_Equalize_Correction",
        ("MULTIPLY", "ADD"),
        conf["HiStitch"]["HiStitch_Equalize_Correction"],
    )

    util.conf_check_count(
        "HiStitch_Balance_Processing",
        14,
        "CCD",
        conf["HiStitch"]["HiStitch_Balance_Processing"],
    )

    util.conf_check_count(
        "HiStitch_Control_Channel",
        14,
        "CCD",
        conf["HiStitch"]["HiStitch_Control_Channel"],
    )

    util.conf_check_count(
        "HiStitch_Balance_Bin_DarkPixel_STD",
        5,
        "bin mode",
        conf["HiStitch"]["HiStitch_Balance_Bin_DarkPixel_STD"],
    )

    util.conf_check_count(
        "HiStitch_Balance_Bin_Mask_STD",
        5,
        "bin mode",
        conf["HiStitch"]["HiStitch_Balance_Bin_Mask_STD"],
    )

    util.conf_check_bounds(
        "HiStitch_Normalization_Minimum",
        (-16384.0, 16364.0),
        conf["HiStitch"]["HiStitch_Normalization_Minimum"],
    )

    util.conf_check_bounds(
        "HiStitch_Normalization_Maximum",
        (-16384.0, 16364.0),
        conf["HiStitch"]["HiStitch_Normalization_Maximum"],
    )

    util.conf_check_bounds(
        "HiStitch_Minimum_Percent",
        (0, 2),
        conf["HiStitch"]["HiStitch_Minimum_Percent"],
    )
    return


if __name__ == "__main__":
    main()
